#include <string>
#include <iostream>

#include "DemoUnit.hh"

using namespace std;

DemoUnit::DemoUnit(Engine &rTEngine) :
  rEngine(rTEngine)
{}

DemoUnit::~DemoUnit(void)
{}

bool DemoUnit::Init(void)
{
  string cFile = "teapot.off";
  if( !teapotData.LoadFile(cFile) )
    {
      cout << "failed to load scene data from file: " << cFile << endl;
      return false;
    };

  /*
  float minX, maxX;
  float minY, maxY;
  float minZ, maxZ;
  teapotData.FindExtents(minX, maxX, minY, maxY, minZ, maxZ);
  cout << "X extents: " << minX << ',' << maxX << " Y extents: " << minY << ',' << maxY
       << " Z extents: " << minZ << ',' << maxZ << endl;
  */

  cFile = "textured.off";
  if( !texturedData.LoadFile(cFile) ) 
    {
      cout << "failed to load scene from file: " << cFile << endl;
      return false;
    };

  return true;
}

void DemoUnit::TeapotDisplaySingle(float yRotAngle)
{
  rEngine.SetCTexture( 0 );
  
  rEngine.ClearTForm();

  rEngine.PushTForm( CreateRotY( yRotAngle*(M_PI*2.00)/360.00) );
  
  rEngine.PushTForm( CreateScale(80.0) );
  rEngine.PushTForm( CreateRevY() );
  rEngine.PushTForm( CreateOffset(320.0, 350.0, 0.00) );
  
  rEngine.ClearDB();
  rEngine.ClearFB();

  TexC junk;

  for(unsigned ii = 0; ii < teapotData.faceVec.size(); ii++)
    {
      TCoord c1, c2, c3;
      teapotData.GetFace(ii, c1,c2,c3);
      
      rEngine.TriFace(c1,c2,c3, junk,junk,junk);
    };

  rEngine.Flush();
}

void DemoUnit::TeapotRotateAboutY(void)
{
  for(unsigned ii = 0; ii < 360; ii++ )
    {
      TeapotDisplaySingle((float)ii);
    };
}

void DemoUnit::TextureDemo(void)
{
  unsigned cNum = 1;
  list<Texture*>::const_iterator tPlace = texturedData.texList.begin();
  while( tPlace != texturedData.texList.end() && cNum < 9 )
    {
      rEngine.UpLoadTexture( *(*tPlace), cNum);

      tPlace++;
      cNum++;
    };

  /* a debugging test (with a modified display XEDriver::SetCTexture)
  for(unsigned ii = 1; ii < cNum; ii++)
    {
      rEngine.SetCTexture( ii );
      sleep(5);
    };
  */

  for(unsigned ii = 0; ii < 360; ii++ )
    {
      rEngine.ClearTForm();
      rEngine.PushTForm( CreateOffset(-0.5,-0.5,-0.5) );
      rEngine.PushTForm( CreateRotZ(M_PI) );

      rEngine.PushTForm( CreateRotX( (float)ii*(M_PI*2.00)/360.00) );
      rEngine.PushTForm( CreateRotY( 2.0*(float)ii*(M_PI*2.00)/360.00) );
      rEngine.PushTForm( CreateRotZ( (float)ii*(M_PI*2.00)/360.00) );

      rEngine.PushTForm( CreateScale(250.00) );
      rEngine.PushTForm( CreateRevY() );
      rEngine.PushTForm( CreateOffset(320.0, 250.0, 0.00) );
      
      rEngine.ClearDB();
      rEngine.ClearFB();
      
      list<ExtFaceData>::const_iterator cPlace = texturedData.faceList.begin();
      while( cPlace != texturedData.faceList.end() )
	{
	  const ExtFaceData &rFace = *cPlace;

	  TCoord c1, c2, c3;
	  c1 = rFace.c1;
	  c2 = rFace.c2;
	  c3 = rFace.c3;
	  TexC t1(rFace.c1.texX, rFace.c1.texY), t2(rFace.c2.texX, rFace.c2.texY), t3(rFace.c3.texX, rFace.c3.texY);
	  
	  rEngine.SetCTexture( rFace.textureID );
	  rEngine.TriFace(c1,c2,c3, t1, t2, t3);

	  cPlace++;
	};
      
      rEngine.Flush();
    };
}
