#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <iostream>

#include "Engine.hh"

using namespace std;

Engine::Engine(EDriver &rDriver) :
  pDriver(&rDriver)
{}

Engine::~Engine(void)
{}

void Engine::TriFace(TCoord rC1, TCoord rC2, TCoord rC3,
		     TexC t1, TexC t2, TexC t3)
{
  rC1 = tForm*rC1;
  rC2 = tForm*rC2;
  rC3 = tForm*rC3;

  TCoord normal( (rC2 - rC1)%(rC3 - rC1) );
  normal.Normalize();

  rC1.z *= 40.00;
  rC2.z *= 40.00;
  rC3.z *= 40.00;

  TCoord lightVec(0.00,0.00, 1.0);
  float intensity = normal*lightVec;
  unsigned long cVal = (unsigned long)rint(intensity*255.00);
  if(cVal > 255) cVal = 255;

  //EPIXEL color = cVal<<EPIXEL_RED_OFFSET | cVal<<EPIXEL_GREEN_OFFSET | cVal<<EPIXEL_BLUE_OFFSET;
  EPIXEL color =  cVal<<EPIXEL_GREEN_OFFSET | cVal<<EPIXEL_BLUE_OFFSET;

  //I'll just cheat and use the lightVec as a back face test.
  if(intensity < 0.00 ) return;

  long x1 = RI(rC1.x);
  long y1 = RI(rC1.y);
  long z1 = RI(rC1.z);

  long x2 = RI(rC2.x);
  long y2 = RI(rC2.y);
  long z2 = RI(rC2.z);

  long x3 = RI(rC3.x);
  long y3 = RI(rC3.y);
  long z3 = RI(rC3.z);

  /*
  cout << x1 << " " << y1 << " " << z1 << endl;
  cout << x2 << " " << y2 << " " << z2 << endl;
  cout << x3 << " " << y3 << " " << z3 << endl;
  cout << endl;
  */

  //cout << "before: " << y1 << " " << y2 << " " << y3 << endl;

  long temp;
  TexC tTmp;

  //sort the values according to increasing y's
  if( y1 >= y2 && y1 > y3) //exchanging y1 and y3
    {
      temp = x3; x3 = x1; x1 = temp;
      temp = y3; y3 = y1; y1 = temp;
      temp = z3; z3 = z1; z1 = temp;
      tTmp = t3; t3 = t1; t1 = tTmp;

      if( y1 > y2) //exchanging old y3 and y2
	{
	  temp = x2; x2 = x1; x1 = temp;
	  temp = y2; y2 = y1; y1 = temp;
	  temp = z2; z2 = z1; z1 = temp;
	  tTmp = t2; t2 = t1; t1 = tTmp;
	};
    }else
      if( y1 > y2)
	{
	  temp = x2; x2 = x1; x1 = temp;
	  temp = y2; y2 = y1; y1 = temp;
	  temp = z2; z2 = z1; z1 = temp;
	  tTmp = t2; t2 = t1; t1 = tTmp;
	}else
	  if( y2 > y3)
	    {
	      temp = x2; x2 = x3; x3 = temp;
	      temp = y2; y2 = y3; y3 = temp;
	      temp = z2; z2 = z3; z3 = temp;
	      tTmp = t2; t2 = t3; t3 = tTmp;
	      if(y1 > y2)
		{
		  temp = x2; x2 = x1; x1 = temp;
		  temp = y2; y2 = y1; y1 = temp;
		  temp = z2; z2 = z1; z1 = temp;
		  tTmp = t2; t2 = t1; t1 = tTmp;
		};
	    };

  //cout << "after: " << y1 << " " << y2 << " " << y3 << endl;

  if( y1 == y2 && y2 == y3 ) return;

  if( y1 > y2 || y1 > y3) assert(false);
  if( y2 > y3) assert(false);

  if( y1 == y2 )
    {
      if( x1 < x2 )
	FlatBottomRast(x1,z1,t1, x2,z2,t2, y1, x3,y3,z3,t3, color);
      else
	FlatBottomRast(x2,z2,t2, x1,z1,t1, y1, x3,y3,z3,t3, color);
      return;
    };

  if( y2 == y3 )
    {
      if( x2 < x3 )
	FlatTopRast(x1,y1,z1,t1, x2,z2,t2, x3,z3,t3, y2, color);
      else
	FlatTopRast(x1,y1,z1,t1, x3,z3,t3, x2,z2,t2, y2, color);
      return;
    };

  if( (y3-y1)*(x2-x1) - (y2 - y1)*(x3-x1) > 0)
    {
      //cout << "right broken rast" << endl;
      RightBrokenRast(x1,y1,z1,t1, x2,y2,z2,t2, x3,y3,z3,t3, color);
      return;
    };

  //cout << "left broken rast " << endl;
  LeftBrokenRast(x1,y1,z1,t1, x2,y2,z2,t2, x3,y3,z3,t3, color);
  return;
}

void Engine::FlatBottomRast(int blX, int blZ, TexC blTC, int brX, int brZ, TexC brTC, int bY, int tX, int tY, int tZ, TexC tTC,
			    EPIXEL pixColor)
{
  assert( tY > bY);
  /*
  cout << "Engine::FlatBottomRast" << endl;
  cout << "bottom leftX: " << blX << '\n'
       << "bottom left TC: " << blTC << '\n'
       << "bottom rightX: " << brX << '\n'
       << "bottom right TC: " << brTC << '\n'
       << "bottom Y: " << bY << '\n'
       << "topX: " << tX << '\n'
       << "topY: " << tY << '\n'
       << "top texture Coord: " << tTC << endl;
  */
  float scale = 1.00/(float)(tY - bY);

  TexC leftTCInc( scale*(tTC - blTC) );
  TexC rightTCInc( scale*(tTC - brTC) );

  TexC leftTC(blTC), rightTC(brTC);
  
  for(int cY = bY; cY < tY; cY++)
    {
      int startX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - blX)) + blX );
      int endX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - brX)) + brX );

      int startZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - blZ)) + blZ );
      int endZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - brZ)) + brZ );

      scale = 1.00/(float)(endX - startX);
      TexC incTC( scale*(rightTC - leftTC) );
      //if( cID == 2 )
      //cout << leftTC << "    " << rightTC <<  "     " << startX << "   " << endX << "      ";
      pDriver->PolyScanLine(startX, startZ, endX, endZ, cY, leftTC, incTC, pixColor);

      leftTC += leftTCInc;
      rightTC += rightTCInc;
    };
}

void Engine::FlatTopRast(int bX, int bY, int bZ, TexC bTC, int tlX, int tlZ, TexC tlTC, int trX, int trZ, TexC trTC, int tY, 
			 EPIXEL pixColor)
{
  //cout << "Engine::FlatTopRast" << endl;
  assert( tY > bY);

  float scale = 1.00/(float)(tY - bY);
  TexC leftTCInc( scale*(tlTC - bTC) );
  TexC rightTCInc( scale*(trTC - bTC) );

  TexC leftTC(bTC), rightTC(bTC);

  for(int cY = bY; cY < tY; cY++)
    {
      int startX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tlX - bX)) + bX );
      int endX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(trX - bX)) + bX );

      int startZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tlZ - bZ)) + bZ );
      int endZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(trZ - bZ)) + bZ );

      scale = 1.00/(float)(endX - startX);
      TexC incTC( scale*(rightTC - leftTC) );

      pDriver->PolyScanLine(startX, startZ, endX, endZ, cY, leftTC, incTC, pixColor);

      leftTC += leftTCInc;
      rightTC += rightTCInc;
   };
}

void Engine::LeftBrokenRast(int bX, int bY, int bZ, TexC bTC, int lBrknX, int lBrknY, int lBrknZ, TexC lBrknTC, int tX, int tY, int tZ, TexC tTC,
			    EPIXEL pixColor)
{
  assert(lBrknY > bY);
  assert(tY > lBrknY);

  TexC leftTCInc( (1.00/(float)(lBrknY - bY) )*(lBrknTC - bTC) );
  TexC rightTCInc( (1.00/(float)(tY-bY) )*(tTC - bTC) );

  TexC leftTC(bTC), rightTC(bTC);

  for(int cY = bY; cY < lBrknY; cY++)
    {
      int startX = RI( ((float)(cY - bY)/(float)(lBrknY - bY))*((float)(lBrknX - bX)) + bX );
      int endX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - bX)) + bX );

      int startZ = RI( ((float)(cY - bY)/(float)(lBrknY - bY))*((float)(lBrknZ - bZ)) + bZ );
      int endZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - bZ)) + bZ );
 
      float scale = 1.00/(float)(endX - startX);
      TexC incTC( scale*(rightTC - leftTC) );

      pDriver->PolyScanLine(startX, startZ, endX, endZ, cY, leftTC, incTC,  pixColor);

      leftTC += leftTCInc;
      rightTC += rightTCInc;
    };

  leftTC = lBrknTC;
  leftTCInc = ( (1.00/(float)(tY - lBrknY) )*(tTC - lBrknTC) );

  for(int cY = lBrknY; cY < tY; cY++)
    {
      int startX = RI( ((float)(cY - lBrknY)/(float)(tY - lBrknY))*((float)(tX - lBrknX)) + lBrknX );
      int endX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - bX)) + bX );

      int startZ = RI( ((float)(cY - lBrknY)/(float)(tY - lBrknY))*((float)(tZ - lBrknZ)) + lBrknZ );
      int endZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - bZ)) + bZ );

      float scale = 1.00/(float)(endX - startX);
      TexC incTC( scale*(rightTC - leftTC) );

      pDriver->PolyScanLine(startX, startZ, endX, endZ, cY, leftTC, incTC, pixColor);

      leftTC += leftTCInc;
      rightTC += rightTCInc;
    };
};

void Engine::RightBrokenRast(int bX, int bY, int bZ, TexC bTC, int rBrknX, int rBrknY, int rBrknZ, TexC rBrknTC, int tX, int tY, int tZ, TexC tTC,
			     EPIXEL pixColor)
{
  assert(rBrknY > bY);
  assert(tY > rBrknY);

  TexC leftTCInc( (1.00/(float)(tY-bY) )*(tTC - bTC) );
  TexC rightTCInc( (1.00/(float)(rBrknY - bY) )*(rBrknTC - bTC) );

  TexC leftTC(bTC), rightTC(bTC);

  for(int cY = bY; cY < rBrknY; cY++)
    {
      int startX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - bX)) + bX );
      int endX = RI( ((float)(cY - bY)/(float)(rBrknY - bY))*((float)(rBrknX - bX)) + bX );

      int startZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - bZ)) + bZ );
      int endZ = RI( ((float)(cY - bY)/(float)(rBrknY - bY))*((float)(rBrknZ - bZ)) + bZ );

      float scale = 1.00/(float)(endX - startX);
      TexC incTC( scale*(rightTC - leftTC) );

      pDriver->PolyScanLine(startX, startZ, endX, endZ, cY, leftTC, incTC, pixColor);
      
      leftTC += leftTCInc;
      rightTC += rightTCInc;
    };

  rightTC = rBrknTC;
  rightTCInc = ( (1.00/(float)(tY - rBrknY) )*(tTC - rBrknTC) );

  for(int cY = rBrknY; cY < tY; cY++)
    {
      int startX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - bX)) + bX );
      int endX = RI( ((float)(cY - rBrknY)/(float)(tY - rBrknY))*((float)(tX - rBrknX)) + rBrknX );

      int startZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - bZ)) + bZ );
      int endZ = RI( ((float)(cY - rBrknY)/(float)(tY - rBrknY))*((float)(tZ - rBrknZ)) + rBrknZ );

      float scale = 1.00/(float)(endX - startX);
      TexC incTC( scale*(rightTC - leftTC) );

      pDriver->PolyScanLine(startX, startZ, endX, endZ, cY, leftTC, incTC,  pixColor);
       
      leftTC += leftTCInc;
      rightTC += rightTCInc;
   };
};

void Engine::ClearFB(void)
{
  pDriver->ClearFB(0);
}
 
void Engine::ClearDB(void)
{
  pDriver->ClearDB(-10000000);
}

void Engine::UpLoadTexture(Texture &rTexture, unsigned textureID)
{
  pDriver->UpLoadTexture(rTexture, textureID);
}

void Engine::SetCTexture(unsigned cTextureID)
{
  pDriver->SetCTexture(cTextureID);
}

/*
void Engine::Test(void)
{
  PolyScanLine( 50, 100, 200, -100, 240,
	    frameBuffer, depthBuffer);

  FlatBottomRast(100, 300, 100, 200, 200, frameBuffer, depthBuffer);

  FlatTopRast(400, 0, 300, 500, 100, frameBuffer, depthBuffer);

  LeftBrokenRast( 200, 250, 50, 350, 300, 400, frameBuffer, depthBuffer);

  RightBrokenRast( 400, 250, 550, 350, 500, 400, frameBuffer, depthBuffer);
};
*/

void Engine::PushTForm(TForm addTForm)
{
  tForm = addTForm*tForm;
}

void Engine::ClearTForm(void)
{
  tForm = CreateIdentity();
}

void Engine::Flush(void)
{
  pDriver->Flush();
}
