#include <assert.h>
#include <fstream>
#include <math.h>

#include "OffData.hh"

using namespace std;

// ***********************

FaceData::FaceData(void)
{}

FaceData::FaceData(unsigned tC1, unsigned tC2, unsigned tC3) :
  c1(tC1), c2(tC2), c3(tC3)
{}

FaceData::FaceData(const FaceData &rhs) :
  c1(rhs.c1), c2(rhs.c2), c3(rhs.c3)
{}
  
FaceData::~FaceData(void)
{}

const FaceData& FaceData::operator=(const FaceData &rhs)
{
  c1 = rhs.c1;
  c2 = rhs.c2;
  c3 = rhs.c3;

  return *this;
}

// **********************

OffData::OffData(void)
{}

OffData::~OffData(void)
{}

bool OffData::LoadFile(const string &rFileName)
{
  ifstream fin( rFileName.c_str() );
  if(!fin) return false;

  enum F_STAT{F_HEADER, F_DESC, F_COORD, F_FACES};
  F_STAT cStat = F_HEADER;

  unsigned verticesRead = 0;
  unsigned facesRead = 0;

  char lineBuffer[512];
  while(true)
    {
      fin.getline(lineBuffer, 511);
      lineBuffer[511] = '\0';

      if( cStat == F_HEADER )
	{
	  if( string(lineBuffer, 3) != "OFF" ) return false;
	  cStat = F_DESC;
	  continue;
	};
  
      if( cStat == F_DESC )
	{
	  const char *pStart = lineBuffer;
	  char *pTail = 0;
	  unsigned numVertices = strtoul(pStart, &pTail, 10);
	  if(pTail == pStart) return false;

	  pStart = ++pTail;
	  unsigned numFaces = strtoul(pStart, &pTail, 10);
	  if(pTail == pStart) return false;

	  pStart = ++pTail;
	  unsigned numEdges = strtoul(pStart, &pTail, 10);
	  if(pTail == pStart) return false;

	  if( numVertices == 0 || numFaces == 0 ) return false;

	  coordVec.resize(numVertices);
	  faceVec.resize(numFaces);

	  cStat = F_COORD;
	  continue;
	};
  
      if( cStat == F_COORD )
	{
	  const char *pStart = lineBuffer;
	  char *pTail = 0;

	  float xVal = strtod(pStart, &pTail);
	  if( pTail == pStart ) return false;

	  pStart = ++pTail;
	  float yVal = strtod(pStart, &pTail);
	  if( pTail == pStart) return false;

	  pStart = ++pTail;
	  float zVal = strtod(pStart, &pTail);
	  if( pTail == pStart) return false;

	  if( verticesRead >= coordVec.size() ) return false; //useless little test

	  coordVec[verticesRead] = TCoord(xVal, yVal, zVal);
	  verticesRead++;

	  if( verticesRead == coordVec.size() )
	    cStat = F_FACES;
	    
	  continue;
	};

      if( cStat == F_FACES )
	{
	  const char *pStart = lineBuffer;
	  char *pTail = 0;
	  unsigned numVertices = strtoul(pStart, &pTail, 10);
	  if(pTail == pStart) return false;

	  if(numVertices != 3) return false;

	  pStart = ++pTail;
	  unsigned c1 = strtoul(pStart, &pTail, 10);
	  if(pTail == pStart) return false;

	  pStart = ++pTail;
	  unsigned c2 = strtoul(pStart, &pTail, 10);
	  if(pTail == pStart) return false;

	  pStart = ++pTail;
	  unsigned c3 = strtoul(pStart, &pTail, 10);
	  if(pTail == pStart) return false;

	  if( c1 >= coordVec.size() ) return false;
	  if( c2 >= coordVec.size() ) return false;
	  if( c3 >= coordVec.size() ) return false;

	  faceVec[facesRead] = FaceData(c1, c2, c3);
	  facesRead++;

	  if( facesRead == faceVec.size() )
	    {
	      fin.close();
	      return true;
	    };

	  continue;
	};
    };

  return false;
}

bool OffData::GetFace(unsigned num, TCoord &rOne, TCoord &rTwo, TCoord &rThree)
{
  if(num > faceVec.size() ) return false;

  FaceData &rFD = faceVec[num];

  if(rFD.c1 >= coordVec.size() ) return false;
  rOne = coordVec[rFD.c1];

  if(rFD.c2 >= coordVec.size() ) return false;
  rTwo = coordVec[rFD.c2];

  if(rFD.c3 >= coordVec.size() ) return false;
  rThree = coordVec[rFD.c3];

  return true;
}

void OffData::FindExtents(float &rMinX, float &rMaxX, float &rMinY, float &rMaxY,
			  float &rMinZ, float &rMaxZ)
{
  assert( !coordVec.empty() );

  rMinX = rMaxX = coordVec[0].x;
  rMinY = rMaxY = coordVec[0].y;
  rMinZ = rMaxZ = coordVec[0].z;

  for(unsigned ii = 1; ii < coordVec.size(); ii++)
    {
      TCoord &rCoord = coordVec[ii];

      if( rCoord.x > rMaxX )
	rMaxX = rCoord.x;
      else
	if( rCoord.x < rMinX)
	  rMinX = rCoord.x;

      if( rCoord.y > rMaxY )
	rMaxY = rCoord.y;
      else
	if( rCoord.y < rMinY)
	  rMinY = rCoord.y;

      if( rCoord.z > rMaxZ )
	rMaxZ = rCoord.z;
      else
	if( rCoord.z < rMinZ)
	  rMinZ = rCoord.z;
    };
}

