#include <fstream>
#include <iostream>

#include "Texture.hh"

using namespace std;

TexC::TexC(void)
{}

TexC::TexC(float tX, float tY) :
  x(tX), y(tY)
{}

TexC::~TexC(void)
{}

std::ostream& operator<<(std::ostream &out, TexC texC)
{
  out << texC.x << ',' << texC.y;
  return out;
}

// *************************


Texture::Texture(void) :
  texWidth(0), texHeight(0), pPixelData(0)
{}

Texture::~Texture(void)
{
  delete [] pPixelData;
  pPixelData = 0;
}

Texture& Texture::operator=(const Texture &rhs)
{
  delete [] pPixelData;
  pPixelData = 0;

  texWidth = rhs.texWidth;
  texHeight = rhs.texHeight;

  pPixelData = new EPIXEL[texWidth*texHeight];
  memcpy(pPixelData, rhs.pPixelData, sizeof(EPIXEL)*texWidth*texHeight);

  return *this;
}

bool Texture::FillTexture(const std::string &rFileName)
{
  ifstream fin( rFileName.c_str() );
  if( !fin ) {
    cerr << "Texture::FillTexture failed to open file: " << rFileName << endl;
    return false;
  };

  char lineBuffer[256];
  fin.getline(lineBuffer, 255);
  lineBuffer[255] = '\0';

  if( string(lineBuffer, 2) != "P3" )
    {
      cerr << "Texture::FillTexture file does not appear to be of expected type" << endl;
      return false;
    };

  //hack. discard a known comment line.
  fin.getline(lineBuffer, 255);
  lineBuffer[255] = '\0';

  fin >> texWidth;
  fin >> texHeight;
  
  if( texWidth != 256 || texHeight != 256 )
    {
      cerr << "unexpected texture size" << endl;
      return false;
    };

  delete [] pPixelData;
  pPixelData = new EPIXEL[texWidth*texHeight];

  unsigned maxSample;
  fin >> maxSample;

  for(unsigned cY = 0; cY < texHeight; cY++)
    for(unsigned cX = 0; cX < texWidth; cX++)
      {
	EPIXEL red, green, blue;

	fin >> red;
	fin >> green;
	fin >> blue;

	EPIXEL value = red<<EPIXEL_RED_OFFSET | green<<EPIXEL_GREEN_OFFSET | blue<<EPIXEL_BLUE_OFFSET;
	pPixelData[cY*texWidth + cX] = value;

	if( !fin)
	  {
	    cerr << "read error" << endl;
	    return false;
	  };
      };

  return true;
}
