#ifndef TVCDRIVER_INCLUDED
#define TVCDRIVER_INCLUDED

#include <vector>

#include "EPPPort.hh"
#include "EDriver.hh"

//this should inherit from a class that only has virtual void functions
class TvcDriver : public EDriver
{
public:
  TvcDriver(EPPPort &rTPort);
  virtual ~TvcDriver(void);

  void DumpRegisters(void); //does not use cache
  void DumpLocalRegisterCache(void);
  void ClearRegisters(void); //does not use cache
  void PrintStatus(void); //does not use cache

  void TestRam(void); //not implemented

  void Blank(unsigned char pixValue = 0);

  void DrawPixel(int xPos, int yPos, unsigned char pixValue);
  void DrawLine(int xStart, int yStart, int xEnd, int yEnd, unsigned char pixValue);
  void ReadValue(unsigned long fbAddress, unsigned long &rValue);
  void WriteValue(unsigned long fbAddress, unsigned long newValue);
  void SetValue(unsigned long fbAddress, unsigned long numBytes, unsigned long newValue);

  // ***************** start EDriver overloads ******************
  virtual void PolyScanLine(unsigned sX, long startZ, unsigned eX, long endZ, unsigned yValue, TexC texStart, TexC texInc, EPIXEL pixValue);
  virtual void ClearFB(EPIXEL pixValue);
  virtual void ClearDB(long depthVal);
  virtual void Flush(void);
  virtual void UpLoadTexture(Texture &rTexture, unsigned textureID);
  virtual void SetCTexture(unsigned cTextureID);
  // ****************** end EDriver overloads *******************

protected:
  unsigned char EPixelToHWPixel(EPIXEL pixValue);
  unsigned long HWPixelToBussWidth(unsigned char hwPixel);

  void DirtyRegisterCache(void); //force a write to all registers to sync up local and remote
  //we cache the remote contents of the data registers so we only write the
  //registers that have changed.
  void FlushRegisters(void);
  std::vector<unsigned char> localDataRegisters;
  std::vector<unsigned char> remoteDataRegisters;
  std::vector<bool> remoteDataValid;
  //unsigned char remoteAddress; we don't cache this right now.
  void Write32(unsigned long value, unsigned regOffset); //writes into the cache.

  void WaitOnBW(void); //poll for completion of the block writer
  void WaitOnBR(void); //same for block reader
  void WaitOnBS(void); //same for block set
  void WaitOnPS(void); //same for poly scanline

  EPPPort &rPort;
};

#endif //TVCDRIVER_INCLUDED
