#ifndef XEDRIVER_INCLUDED
#define XEDRIVER_INCLUDED

#include <libImg/SimpleImg.hh>
#include <libImg/U32RGBAPixel.hh>
#include <libImg/PXMUtil.hh>
#include <libImg/XImageWindow.hh>

#include "EDriver.hh"

class DepthPixel
{
public:
  DepthPixel(void) : value(0)
  {};
  DepthPixel(long tValue) : value(tValue) {};
  DepthPixel(const DepthPixel &rhs) : value(rhs.value) {};
  ~DepthPixel(void){};

  const DepthPixel& operator=(const DepthPixel &rhs)
    {
      value = rhs.value;
      return *this;
    };

  long value;
};

class XEDriver : public EDriver
{
public:
  XEDriver(void);
  virtual ~XEDriver(void);

  virtual void CreateRenderContext(void);
  virtual void DeleteRenderContext(void);

  virtual void PolyScanLine(unsigned sX, long startZ, unsigned eX, long endZ, unsigned yValue,
			    TexC texStart, TexC texInc, EPIXEL pixValue);

  virtual void ClearFB(EPIXEL pixValue);
  virtual void ClearDB(long depthVal);

  virtual void Flush(void);

  virtual void UpLoadTexture(Texture &rTexture, unsigned textureID);
  virtual void SetCTexture(unsigned cTextureID);

protected:
  XImageWindow xImgWindow;

  unsigned currentTex; //zero is no texture is current
  Texture texArray[8];

  SimpleImg<unsigned, U32RGBAPixel> frameBuffer;
  SimpleImg<unsigned, DepthPixel> depthBuffer;
};

#endif //XEDRIVER_INCLUDED
