library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity block_reader is
	Port(
   --memory user access buss 
   mu_buss_address 	   : in  std_logic_vector(11 downto 0);
   mu_buss_data		   : inout std_logic_vector(31 downto 0);
   mu_buss_write	   	: in  std_logic; --if high the MCU is going to write to a mu
   --if high the MCU expects the selected MU to present its data on the mu_buss_data or to
   --copy data from the mu_buss_data into its internal cache/registers.
   mu_buss_strobe			: in  std_logic;
   
   block_reader_irq		: out std_logic;
   block_reader_select	: in  std_logic;

   --blah		 
   tvc_clock		      : in	std_logic;
   tvc_reset		      : in 	std_logic;

    --interface to the eppcu
   epp_fb_address 	   : in  std_logic_vector(31 downto 0);
   epp_fb_data		      : out std_logic_vector(31 downto 0);
   epp_start_strobe	   : in  std_logic;
   epp_idle			      : out std_logic
   );
end block_reader;

architecture Behavioral of block_reader is
	signal fb_address_reg 	: STD_LOGIC_VECTOR(31 downto 0) := (others => '0');
	--signal fb_data_reg		: STD_LOGIC_VECTOR(31 downto 0);

	signal selected_once    : STD_LOGIC;

	TYPE PR_INT_STATE is ( PR_IDLE, PR_MCU_DRIVEN);
	signal pr_state         : PR_INT_STATE := PR_IDLE;

	signal mu_buss_data_in  : std_logic_vector(31 downto 0);
	signal mu_buss_data_out : std_logic_vector(31 downto 0);

begin
	process (pr_state)
	begin
		if pr_state = PR_IDLE then
			epp_idle <= '1';
		else
			epp_idle <= '0';
		end if;
	end process;

	mu_buss_data_in <= mu_buss_data;
	process(mu_buss_data_out, block_reader_select, mu_buss_strobe, mu_buss_write)
	begin
		if block_reader_select = '1' and mu_buss_strobe = '1' and mu_buss_write = '0' then
			mu_buss_data <= mu_buss_data_out;
		else
			mu_buss_data <= (others => 'Z');
		end if;
	end process;

	Process (tvc_clock)
	begin
	if tvc_clock = '1' and tvc_clock'event then
	if tvc_reset = '1' then
		block_reader_irq <= '0';
		pr_state <= PR_IDLE;
	else --not reset
	   case pr_state is
	   when PR_IDLE =>
	   	if epp_start_strobe = '1' then
	   		fb_address_reg <= epp_fb_address;
				block_reader_irq <= '1';
				pr_state <= PR_MCU_DRIVEN;
				selected_once <= '0';
		end if;
		when PR_MCU_DRIVEN =>
			if block_reader_select = '0' and selected_once = '1' then
				pr_state <= PR_IDLE;
				block_reader_irq <= '0';
			elsif block_reader_select = '1' then
				selected_once <= '1';
				if mu_buss_address = X"FFC" and mu_buss_write = '0' and mu_buss_strobe = '1' then
					mu_buss_data_out <= X"00" & X"000004";
				elsif mu_buss_address = X"FFD" and mu_buss_write = '0' and mu_buss_strobe = '1' then
					mu_buss_data_out <= fb_address_reg;
				elsif mu_buss_address = X"000" and mu_buss_write = '1' and mu_buss_strobe = '1' then
					epp_fb_data <= mu_buss_data_in;
				end if;
			end if;
		end case;
	end if; --reset
	end if; --clock
	end process;
end Behavioral;
