----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    03:03:41 09/16/2006 
-- Design Name: 
-- Module Name:    block_set - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

--HACK!!! this must break the transfer up into smaller chunks to prevent pix_pusher data starvation.
--right now this is handled in the driver, but this is quite tacky.

entity block_set is
	Port(
		--memory user access buss 
		mu_buss_address 	: in std_logic_vector(11 downto 0);
		mu_buss_data		: inout std_logic_vector(31 downto 0);
		mu_buss_write		: in  std_logic;
		mu_buss_strobe		: in  std_logic;
		
		block_set_irq		: out std_logic;
		block_set_select	: in std_logic;

		tvc_clock			: in std_logic;
		tvc_reset			: in std_logic;

		 --interface to the eppcu
		epp_fb_address 	: in std_logic_vector(31 downto 2);
		epp_num_bytes		: in std_logic_vector(23 downto 2);
		epp_fb_value		: in std_logic_vector(31 downto 0);
		epp_start_strobe	: in std_logic;
		epp_idle				: out std_logic
	  );
end block_set;

architecture Behavioral of block_set is

	signal fb_address_reg 	: std_logic_vector(31 downto 2);
	signal fb_num_reg			: std_logic_vector(23 downto 2);
	signal fb_value_reg 		: std_logic_vector(31 downto 0);

	signal selected_once : STD_LOGIC;

	TYPE BS_INT_STATE is ( BS_IDLE, BS_MCU_DRIVEN);
	signal bs_state : BS_INT_STATE;

	--signal mu_buss_data_in : std_logic_vector(31 downto 0);
	signal mu_buss_data_out : std_logic_vector(31 downto 0);

	procedure HandleMCUDrivenClock(
		signal bs_state				: inout BS_INT_STATE;
		signal fb_address_reg 		: in std_logic_vector(31 downto 2);
		signal fb_num_reg				: in std_logic_vector(23 downto 2);
		signal fb_value_reg 			: in std_logic_vector(31 downto 0);
		signal selected_once			: inout std_logic;
		signal mu_buss_address 		: in std_logic_vector(11 downto 0);
		signal mu_buss_data_out		: out std_logic_vector(31 downto 0);
		signal mu_buss_write			: in  std_logic;
		signal mu_buss_strobe		: in  std_logic;
		signal block_set_irq 		: out std_logic;
		signal block_set_select 	: in std_logic ) is
	begin
		if block_set_select = '0' and selected_once = '1' then
			bs_state <= BS_IDLE;
			block_set_irq <= '0';
		 elsif block_set_select = '1' then
			selected_once <= '1';
			if mu_buss_address = X"FFC" and mu_buss_write = '0' and mu_buss_strobe = '1' then
				mu_buss_data_out <= X"80" & fb_num_reg & "00";
			elsif mu_buss_address = X"FFD" and mu_buss_write = '0' and mu_buss_strobe = '1' then
				mu_buss_data_out <= fb_address_reg & "00";
			elsif mu_buss_write = '0' and mu_buss_strobe = '1' then
				mu_buss_data_out <= fb_value_reg;
			end if;
		end if;
	end HandleMCUDrivenClock;

begin

	--mu_buss_data_in <= mu_buss_data;

	process (bs_state)
	begin
		if bs_state = BS_IDLE then
			epp_idle <= '1';
		else
			epp_idle <= '0';
		end if;
	end process;

	process(mu_buss_data_out, block_set_select, mu_buss_strobe, mu_buss_write)
	begin
		if block_set_select = '1' and mu_buss_strobe = '1' and mu_buss_write = '0' then
			mu_buss_data <= mu_buss_data_out;
		else
			mu_buss_data <= (others => 'Z');
		end if;
	end process;

	Process (tvc_clock)
	begin
	if tvc_clock = '1' and tvc_clock'event then
	if tvc_reset = '1' then
		--fb_address_reg <= X"0000_000" & "00";
		--fb_num_reg <= X"00_000" & "00";
		--fb_value_reg <= X"0000_0000";
		selected_once <= '0';
		bs_state <= BS_IDLE;

		block_set_irq <= '0';
	else --not reset
	   case bs_state is
	   when BS_IDLE =>
	   	if epp_start_strobe = '1' then
				fb_address_reg <= epp_fb_address;
				fb_num_reg <= epp_num_bytes;
				fb_value_reg <= epp_fb_value;

	   		fb_address_reg <= epp_fb_address;
				block_set_irq <= '1';
				bs_state <= BS_MCU_DRIVEN;
				selected_once <= '0';
			end if;
		when BS_MCU_DRIVEN =>
			HandleMCUDrivenClock(bs_state, fb_address_reg, fb_num_reg, fb_value_reg, selected_once,
						 	mu_buss_address, mu_buss_data_out, mu_buss_write, mu_buss_strobe,
								block_set_irq, block_set_select);
		end case;
	end if; --reset
	end if; --clock
	end process;

end Behavioral;

