library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
--library UNISIM;
--use UNISIM.VComponents.all;

entity block_writer is
	Port(
		--memory user access buss 
		mu_buss_address 	: in std_logic_vector(11 downto 0);
		mu_buss_data		: inout std_logic_vector(31 downto 0);
		mu_buss_write		: in  std_logic; --if high the MCU is going to write to a mu
		 --if high the MCU expects the selected MU to present its data on the mu_buss_data or to
		 --copy data from the mu_buss_data into its internal cache/registers.
		mu_buss_strobe			: in  std_logic;
		
		block_writer_irq		: out std_logic;
		block_writer_select	: in std_logic;

		--blah		 
		 tvc_clock		: in		std_logic;
		 tvc_reset		: in 	std_logic;

		 --interface to the eppcu
		epp_fb_address 	: in std_logic_vector(31 downto 0);
		epp_fb_data		: in std_logic_vector(31 downto 0);
		epp_start_strobe	: in std_logic;
		epp_idle			: out std_logic
		);
end block_writer;

architecture Behavioral of block_writer is
	signal fb_address_reg 	: STD_LOGIC_VECTOR(31 downto 0);
	signal fb_data_reg		: STD_LOGIC_VECTOR(31 downto 0);

	signal selected_once : STD_LOGIC;

	TYPE PW_INT_STATE is ( PW_IDLE, PW_MCU_DRIVEN, PW_COMPLETING );
	signal pw_state : PW_INT_STATE;

 	signal mu_buss_data_out : std_logic_vector(31 downto 0);

	procedure HandleMCUDrivenClock( signal pw_state	: inout PW_INT_STATE;
								signal fb_address_reg 		: in STD_LOGIC_VECTOR(31 downto 0);
								signal fb_data_reg			: in	STD_LOGIC_VECTOR(31 downto 0);
								signal selected_once			: inout std_logic;
								signal mu_buss_address 		: in std_logic_vector(11 downto 0);
								signal mu_buss_data_out		: out std_logic_vector(31 downto 0);
								signal mu_buss_write			: in  std_logic;
								signal mu_buss_strobe		: in  std_logic;
								signal block_writer_irq		: out std_logic;
								signal block_writer_select : in std_logic) is
	begin
		if block_writer_select = '0' and selected_once = '1' then
			pw_state <= PW_COMPLETING;
			block_writer_irq <= '0';
		 elsif block_writer_select = '1' then
		 	selected_once <= '1';
			if mu_buss_address = X"FFC" and mu_buss_write = '0' and mu_buss_strobe = '1' then
				mu_buss_data_out <= X"80" & X"000004";
			elsif mu_buss_address = X"FFD" and mu_buss_write = '0' and mu_buss_strobe = '1' then
				mu_buss_data_out <= fb_address_reg;
			elsif mu_buss_address = X"000" and mu_buss_write = '0' and mu_buss_strobe = '1' then
				mu_buss_data_out <= fb_data_reg;
			end if;
		end if;
	end HandleMCUDrivenClock;

begin
	process(pw_state)
	begin
		if pw_state = PW_IDLE then
			epp_idle <= '1';
		else
			epp_idle <= '0';
		end if;
	end process;

	process(mu_buss_data_out, block_writer_select, mu_buss_strobe, mu_buss_write)
	begin
		if block_writer_select = '1' and mu_buss_strobe = '1' and mu_buss_write = '0' then
			mu_buss_data <= mu_buss_data_out;
		else
			mu_buss_data <= (others => 'Z');
		end if;
	end process;

	Process (tvc_clock)
	begin
	if tvc_clock = '1' and tvc_clock'event then
	if tvc_reset = '1' then
		block_writer_irq <= '0';
		--fb_address_reg <= X"0000_0000";
		--fb_data_reg <= X"0000_0000";
		pw_state <= PW_IDLE;
	else --not reset
		case pw_state is
	  	when PW_IDLE =>
		  	if epp_start_strobe = '1' then
	   		fb_address_reg <= epp_fb_address;
				fb_data_reg <= epp_fb_data;
				--fb_data_reg <= X"1234_5678";
				block_writer_irq <= '1';
				pw_state <= PW_MCU_DRIVEN;
				selected_once <= '0';
			end if;
		when PW_MCU_DRIVEN =>
			HandleMCUDrivenClock(pw_state, fb_address_reg, fb_data_reg, selected_once,
							mu_buss_address, mu_buss_data_out, mu_buss_write, mu_buss_strobe, block_writer_irq, block_writer_select);
		when PW_COMPLETING =>
			pw_state <= PW_IDLE;
		end case;
	end if; --reset
	end if; --clock
	end process;
end Behavioral;
