----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    07:28:44 08/03/2006 
-- Design Name: 
-- Module Name:    mcu_hl - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity mcu_hl is
   Port (
 		--memory user access buss 
		mu_buss_address 	: out std_logic_vector(11 downto 0);
		mu_buss_data		: inout std_logic_vector(31 downto 0);
		mu_buss_write		: out  std_logic; --if high the MCU is going to write to a mu
		--if high the MCU expects the selected MU to present its data on the mu_buss_data or to
		--copy data from the mu_buss_data into its internal cache/registers.
		mu_buss_strobe		: out  std_logic;

		tvc_clock		: in	std_logic;
		tvc_reset		: in 	std_logic;

		mcu_hl_start_strobe	: in std_logic; --the scheduler takes this high after it preps the mu
		mcu_hl_busy				: out std_logic; --we take this high after observing the above go high, and remains high until we are done with the mu.

		--after we initiate a transfer, the mcu controls all timing
		mcu_ll_data_out 			: out std_logic_vector(31 downto 0);
		mcu_ll_data_in				: in std_logic_vector(31 downto 0);
		mcu_ll_address				: out std_logic_vector(31 downto 2);
		mcu_ll_xfer_write			: out	std_logic; --if high we are going to write to the mcu_ll
		mcu_ll_xfer_count			: out std_logic_vector(11 downto 0);
		mcu_ll_data_valid_done 	: in std_logic; --either the mcu is done reading our data or has just written new data to us. high only a single cycle
		mcu_ll_start_strobe 		: out std_logic;
		mcu_ll_idle					: in std_logic --goes high on cycle after start strobe. goes low if transfer complete or an error occurs.
		);
end mcu_hl;

architecture Behavioral of mcu_hl is

	--hackish abstractions that let me change the design around quickly until I figure out what is legal and what is a good idea.
	signal mu_buss_data_in : std_logic_vector(31 downto 0);
	signal mu_buss_data_out : std_logic_vector(31 downto 0);
	signal mu_buss_write_sig : std_logic;
	signal mu_buss_strobe_sig : std_logic;
	
	signal mu_next_address : std_logic_vector(11 downto 0);
	signal mu_next_address_xfer_end_value : std_logic_vector(11 downto 0);

	TYPE MCU_HL_STATE is
	(
	MCU_HL_IDLE,
	MCU_HL_STARTUP,
	MCU_HL_STARTUP_2,
	MCU_HL_STARTUP_VALIDATE,
	MCU_HL_STARTUP_CONFIGURE,
	MCU_HL_READ_ALIGNED_XFER, --four byte aligned start but finish don't care about alignment
	MCU_HL_READ_UNALIGNED_XFER,
	MCU_HL_WRITE_ALIGNED_XFER, --four byte aligned start and finish
	MCU_HL_WRITE_UNALIGNED_XFER,
	MCU_HL_WAITSTATE
	);
	signal hl_state : MCU_HL_STATE;
	signal hl_after_waitstate_state : MCU_HL_STATE;

	--this is the data 'parsed' from the selected mu's request registers.
	signal req_byte_address 	: std_logic_vector(31 downto 0); --the address in ram the mu is having us read/write from.
 	signal req_num_bytes			: std_logic_vector(23 downto 0); --the number of bytes the mu is having us read/write for it.
	signal req_write				: std_logic;

	--values comptuted from the mu's configuration data
	signal xfer_tail_address 	: std_logic_vector(31 downto 0); --is start address plus req_num_bytes i.e. one past the final address
	signal xfer_num_transfers 	: std_logic_vector(11 downto 0);
	signal xfer_first_aligned 	: std_logic; --if doing a write, we need to know if we have to pre-read the first address block and compose it into the mu's data
	signal xfer_final_aligned 	: std_logic; --if doing a write, we need to know if we have to pre-read the last address block and compose it into the mu's data
	signal xfer_alignment 		: std_logic_vector(1 downto 0);	
	
	TYPE READ_ALIGNED_STATE is
	(
	READ_ALIGNED_INIT,
	READ_ALIGNED_BODY
	);
	signal ra_state : READ_ALIGNED_STATE;

	procedure StartReadAligned(
							signal hl_state						: inout MCU_HL_STATE;
							signal ra_state						: inout READ_ALIGNED_STATE;
							signal mu_buss_write_sig			: out std_logic;
							signal mu_next_address				: out std_logic_vector(11 downto 0);
							signal mcu_ll_address				: out std_logic_vector(31 downto 2);
							signal mcu_ll_xfer_write			: out	std_logic;
							signal mcu_ll_xfer_count			: out std_logic_vector(11 downto 0);
							signal mcu_ll_start_strobe_sig	: out std_logic ) is
	begin
		hl_state <= MCU_HL_READ_ALIGNED_XFER;
		ra_state <= READ_ALIGNED_INIT;
		mcu_ll_address <= req_byte_address(31 downto 2);
		mcu_ll_xfer_count <= xfer_num_transfers;
		mcu_ll_xfer_write <= '0';
		mcu_ll_start_strobe_sig <= '1';

		mu_buss_write_sig <= '1'; --only mu buss writes from this point on
		mu_next_address <= X"000";
	end StartReadAligned;

	procedure HandleReadAligned(
							signal hl_state						: inout MCU_HL_STATE;
							signal ra_state						: inout READ_ALIGNED_STATE;
							signal mu_buss_address				: out std_logic_vector(11 downto 0);
							signal mu_buss_data_out				: out std_logic_vector(31 downto 0);
							signal mu_buss_write_sig			: out std_logic;
							signal mu_buss_strobe_sig			: out std_logic;
							signal mu_next_address 				: inout std_logic_vector(11 downto 0);
							signal mcu_ll_data_in				: in std_logic_vector(31 downto 0);
							signal mcu_ll_data_valid_done 	: in std_logic;
							signal mcu_ll_start_strobe_sig	: out	std_logic;
							signal mcu_ll_idle					: in std_logic ) is
	begin
		case ra_state is
		when READ_ALIGNED_INIT =>
			mcu_ll_start_strobe_sig <= '0';
			ra_state <= READ_ALIGNED_BODY;
		when READ_ALIGNED_BODY =>
			if mcu_ll_idle = '1' then
				hl_state <= MCU_HL_IDLE;
				mu_buss_write_sig <= '0';
			else
				if	mcu_ll_data_valid_done = '1' then
					mu_buss_data_out <= mcu_ll_data_in;
					mu_buss_strobe_sig <= '1';
					mu_buss_address <= mu_next_address;
					mu_next_address <= mu_next_address + 1;
				else
					mu_buss_strobe_sig <= '0';
				end if;
			end if;
		end case;
	end HandleReadAligned;

	TYPE READ_UNALIGNED_STATE is
	(
	READ_UNALIGNED_INIT,
	READ_UNALIGNED_HEAD,
	READ_UNALIGNED_BODY,
	READ_UNALIGNED_SINGLE,
	READ_UNALIGNED_COMPLETING
	);
	signal rua_state : READ_UNALIGNED_STATE;

 	--when reading or writing to an unaligned address, we will have to read a initial aligned block.
	signal aligned_dram_data : std_logic_vector(31 downto 0);

	--if alignment != '00' will have to do an initial read into aligned_dram_data before starting HandleXFER
	--on the final transfer we might put a few extra valid data bytes on the mu_buss_data that the mu did not ask for.
	--this is harmless.
	
	procedure StartReadUnAligned( --hack. this can be folded into StartReadAligned
							signal hl_state						: out MCU_HL_STATE;
							signal rua_state						: out READ_UNALIGNED_STATE;
							signal mu_buss_write_sig			: out std_logic;
							signal mu_next_address				: out std_logic_vector(11 downto 0);
							signal mcu_ll_address				: out std_logic_vector(31 downto 2);
							signal mcu_ll_xfer_write			: out	std_logic;
							signal mcu_ll_xfer_count			: out std_logic_vector(11 downto 0);
							signal mcu_ll_start_strobe_sig	: out std_logic ) is
	begin
		hl_state <= MCU_HL_READ_UNALIGNED_XFER;
		rua_state <= READ_UNALIGNED_INIT;
		mcu_ll_address <= req_byte_address(31 downto 2);
		mcu_ll_xfer_count <= xfer_num_transfers;
		mcu_ll_xfer_write <= '0';
		mcu_ll_start_strobe_sig <= '1';

		--mu_buss_strobe_sig <= '0'; --hack. not neccessary.
		mu_buss_write_sig <= '1'; --only mu buss writes from this point on
		--mu_next_address <= X"000"; --done below
	end StartReadUnAligned;

	procedure HandleReadUnAligned(
							signal hl_state								: inout MCU_HL_STATE;
							signal rua_state								: inout READ_UNALIGNED_STATE;
							signal xfer_num_transfers					: in std_logic_vector(11 downto 0);
							signal aligned_dram_data					: inout std_logic_vector(31 downto 0);
							signal mu_buss_address						: out std_logic_vector(11 downto 0);
							signal mu_buss_data_out						: out std_logic_vector(31 downto 0);
							signal mu_buss_write_sig					: out std_logic;
							signal mu_buss_strobe_sig					: out std_logic;
							signal mu_next_address 						: inout std_logic_vector(11 downto 0);
							signal mu_next_address_xfer_end_value 	: inout std_logic_vector(11 downto 0);
							signal mcu_ll_data_in						: in std_logic_vector(31 downto 0);
							signal mcu_ll_data_valid_done 			: in std_logic;
							signal mcu_ll_start_strobe_sig			: out	std_logic;
							signal mcu_ll_idle							: in std_logic ) is
	begin
		case rua_state is
		when READ_UNALIGNED_INIT =>
			mcu_ll_start_strobe_sig <= '0';
			if xfer_num_transfers = X"001" then
				rua_state <= READ_UNALIGNED_SINGLE;
			else
				rua_state <= READ_UNALIGNED_HEAD;
			end if;
		when READ_UNALIGNED_HEAD =>
			if	mcu_ll_data_valid_done = '1' then
				rua_state <= READ_UNALIGNED_BODY;
				aligned_dram_data <= mcu_ll_data_in;
				mu_next_address <= X"000";
				mu_next_address_xfer_end_value <= xfer_num_transfers - 2;
			end if;
		when READ_UNALIGNED_BODY =>
			if	mcu_ll_data_valid_done = '1' then
				case xfer_alignment is
				when "00" =>
					mu_buss_data_out <=  mcu_ll_data_in;
				when "01" =>
					mu_buss_data_out <= aligned_dram_data(23 downto 0) & mcu_ll_data_in(31 downto 24);
					aligned_dram_data <= mcu_ll_data_in;
				when "10" =>
					mu_buss_data_out <= aligned_dram_data(15 downto 0) & mcu_ll_data_in(31 downto 16);
					aligned_dram_data <= mcu_ll_data_in;
				when others => --"11" =>
					mu_buss_data_out <= aligned_dram_data(7 downto 0) & mcu_ll_data_in(31 downto 8);
					aligned_dram_data <= mcu_ll_data_in;
				end case;		

				mu_buss_strobe_sig <= '1';
				mu_buss_address <= mu_next_address;
				if mu_next_address <= mu_next_address_xfer_end_value then
					rua_state <= READ_UNALIGNED_COMPLETING;
				end if;
				
				mu_next_address <= mu_next_address + 1;
			else --mu_ll_data_valid_done
				mu_buss_strobe_sig <= '0';
			end if;
		when READ_UNALIGNED_SINGLE =>
			if	mcu_ll_data_valid_done = '1' then
				case xfer_alignment is
				when "01" =>
					mu_buss_data_out <= mcu_ll_data_in(23 downto 0) & X"00";
				when "10" =>
					mu_buss_data_out <= mcu_ll_data_in(15 downto 0) & X"0000";
				when others =>
					mu_buss_data_out <= mcu_ll_data_in(7 downto 0) & X"000000";
				end case;
				mu_buss_strobe_sig <= '1';
				mu_buss_address <= X"000"; --mu_next_address;
				rua_state <= READ_UNALIGNED_COMPLETING;
			end if;
		when READ_UNALIGNED_COMPLETING =>
			mu_buss_strobe_sig <= '0';
			mu_buss_write_sig <= '0';
			if mcu_ll_idle = '1' then
				hl_state <= MCU_HL_IDLE;
			end if;
		end case;
	end HandleReadUnAligned;
 
	TYPE WRITE_ALIGNED_STATE is
	(
	WRITE_ALIGNED_INIT_DELAY1,
	WRITE_ALIGNED_INIT_DELAY2,
	WRITE_ALIGNED_INIT,
	WRITE_ALIGNED_INIT_DELAY,
	WRITE_ALIGNED_LL_WAIT,
	WRITE_ALIGNED_MU_WAIT
	);
	signal wa_state : WRITE_ALIGNED_STATE;

	procedure StartWriteAligned(
							signal hl_state						: out MCU_HL_STATE;
							signal wa_state						: out WRITE_ALIGNED_STATE;
							signal mu_buss_address 				: out std_logic_vector(11 downto 0);
							signal mu_buss_write_sig			: out std_logic;
							signal mu_buss_strobe_sig 			: out std_logic;
							signal mu_next_address				: out std_logic_vector(11 downto 0);
							signal mcu_ll_address				: out std_logic_vector(31 downto 2);
							signal mcu_ll_xfer_write			: out	std_logic;
							signal mcu_ll_xfer_count			: out std_logic_vector(11 downto 0);
							signal mcu_ll_start_strobe_sig	: out std_logic ) is
	begin
		hl_state <= MCU_HL_WRITE_ALIGNED_XFER;
		wa_state <= WRITE_ALIGNED_INIT_DELAY1;
		mcu_ll_address <= req_byte_address(31 downto 2);
		mcu_ll_xfer_write <= '1';
		mcu_ll_xfer_count <= xfer_num_transfers;
		--mcu_ll_start_strobe_sig <= '1';

		mu_buss_address <= X"000";
		mu_buss_write_sig <= '0'; --only mu buss writes from this point on
		mu_buss_strobe_sig <= '1'; --get the first data available.

		mu_next_address <= X"001";
	end StartWriteAligned;

	procedure HandleWriteAligned(
							signal hl_state						: inout MCU_HL_STATE;
							signal wa_state						: inout WRITE_ALIGNED_STATE;
							signal mu_buss_address				: out std_logic_vector(11 downto 0);
							signal mu_buss_data_in				: in std_logic_vector(31 downto 0);
							signal mu_buss_write_sig			: out std_logic;
							signal mu_buss_strobe_sig			: out std_logic;
							signal mu_next_address				: inout std_logic_vector(11 downto 0);
							signal mcu_ll_data_out 				: out std_logic_vector(31 downto 0);
							signal mcu_ll_data_valid_done 	: in std_logic;
							signal mcu_ll_start_strobe_sig	: out	std_logic;
							signal mcu_ll_idle					: in std_logic ) is
	begin
		case wa_state is
		when WRITE_ALIGNED_INIT_DELAY1 =>
			wa_state <= WRITE_ALIGNED_INIT_DELAY2;
		when WRITE_ALIGNED_INIT_DELAY2 =>
			wa_state <= WRITE_ALIGNED_INIT;
		when WRITE_ALIGNED_INIT =>
			mcu_ll_data_out <= mu_buss_data_in; --make the first transfer data availiable
			mcu_ll_start_strobe_sig <= '1'; --start the full transfer cycle
			wa_state <= WRITE_ALIGNED_INIT_DELAY;
		when WRITE_ALIGNED_INIT_DELAY => --this assumes the mcu_ll won't give us a done signal this clock cycle. pretty safe (it must do dram setup)
			mcu_ll_start_strobe_sig <= '0';
			wa_state <= WRITE_ALIGNED_LL_WAIT;
		when WRITE_ALIGNED_LL_WAIT =>
			if mcu_ll_idle = '1' then --done or error.
				hl_state <= MCU_HL_IDLE; --it will take two clockcycles before the mu_buss_changes
				mu_buss_write_sig <= '0'; --should already be zero.
				mu_buss_strobe_sig <= '0';
			else	
				if	mcu_ll_data_valid_done = '1' then
					mu_buss_address <= mu_next_address;
					mu_buss_strobe_sig <= '1';
					mu_next_address <= mu_next_address + 1;
					wa_state <= WRITE_ALIGNED_MU_WAIT;
				end if;
			end if;
		when WRITE_ALIGNED_MU_WAIT =>
			mcu_ll_data_out <= mu_buss_data_in;
			wa_state <= WRITE_ALIGNED_LL_WAIT;
			mu_buss_strobe_sig <= '0';
		end case;
	end HandleWriteAligned;

	TYPE WRITE_UNALIGNED_STATE is
	(
	--WRITE_UNALIGNED_INIT_DELAY1,
	--WRITE_UNALIGNED_INIT_DELAY2,
	WRITE_UNALIGNED_CLEAR_LL_START_STROBE,
	WRITE_UNALIGNED_READ_SINGLE,
	WRITE_UNALIGNED_READ_SINGLE_WAIT_LL_IDLE,
	WRITE_UNALIGNED_READ_HEAD,
	WRITE_UNALIGNED_READ_HEAD_WAIT_LL_IDLE,
	WRITE_UNALIGNED_READ_TAIL,
	WRITE_UNALIGNED_READ_TAIL_WAIT_LL_IDLE,
	WRITE_UNALIGNED_BODY_STARTUP_DELAY,
	WRITE_UNALIGNED_BODY_STARTUP,
	WRITE_UNALIGNED_BODY,
	WRITE_UNALIGNED_COMPLETING,
	WRITE_UNALIGNED_SINGLE_COMPLETING
	);
	signal wua_state : WRITE_UNALIGNED_STATE;
	signal wua_state_after_clear_start : WRITE_UNALIGNED_STATE;

	--use aligned_dram_data instead of this head_data.
	--signal aligned_write_head_data : std_logic_vector(31 downto 0); --at the start of the memory span
	signal aligned_write_tail_data : std_logic_vector(31 downto 0); --at the end of the memory span
	signal mcu_ll_start_strobe_sig : std_logic;
	signal write_unaligned_final : std_logic;

	procedure SetupUnAlignedHeadXFER(
		signal mcu_ll_address				: out std_logic_vector(31 downto 2);
		signal mcu_ll_xfer_write			: out	std_logic;
		signal mcu_ll_xfer_count			: out std_logic_vector(11 downto 0);
		signal mcu_ll_start_strobe_sig	: out std_logic;
		signal mu_buss_write_sig			: out std_logic ) is
	begin
		mcu_ll_address <= req_byte_address(31 downto 2);
		mcu_ll_xfer_count <= X"001";
		mcu_ll_xfer_write <= '0';
		mcu_ll_start_strobe_sig <= '1';
		mu_buss_write_sig <= '1';
	end SetupUnAlignedHeadXFER;

	procedure SetupUnAlignedTailXFER(
		signal wua_state							: out WRITE_UNALIGNED_STATE;
		signal wua_state_after_clear_start 	: out WRITE_UNALIGNED_STATE;
		signal mcu_ll_address					: out std_logic_vector(31 downto 2);
		signal mcu_ll_xfer_write				: out	std_logic;
		signal mcu_ll_xfer_count				: out std_logic_vector(11 downto 0);
		signal mcu_ll_start_strobe_sig		: out std_logic ) is
	begin
		wua_state <= WRITE_UNALIGNED_CLEAR_LL_START_STROBE;
		wua_state_after_clear_start <= WRITE_UNALIGNED_READ_TAIL;	

		mcu_ll_address <= xfer_tail_address(31 downto 2);
		mcu_ll_xfer_write <= '0';
		mcu_ll_xfer_count <= X"001";
		mcu_ll_start_strobe_sig <= '1';
	end SetupUnAlignedTailXFER;

	procedure StartMUBodyXFER(
		signal mu_buss_write_sig 	: out std_logic;
		signal mu_buss_strobe_sig 	: out std_logic;
		signal mu_buss_address 		: out std_logic_vector(11 downto 0);
		signal mu_next_address 		: out std_logic_vector(11 downto 0) ) is
	begin
		mu_buss_write_sig <= '0'; --only mu buss reads from this point on
		mu_buss_address <= X"000";
		mu_next_address <= X"001";
		mu_buss_strobe_sig <= '1'; --get the first data from the mu ready.
	end StartMUBodyXFER;

	procedure StartWriteUnAligned(
		signal xfer_num_transfers  				: in std_logic_vector(11 downto 0);
		signal xfer_first_aligned					: in std_logic;
		signal mu_buss_write_sig 					: out std_logic;
		signal mu_buss_strobe_sig 					: out std_logic;
		signal mu_buss_address 						: out std_logic_vector(11 downto 0);
		signal hl_state 								: out MCU_HL_STATE;
		signal mu_next_address 						: out std_logic_vector(11 downto 0);
		signal mu_next_address_xfer_end_value 	: out std_logic_vector(11 downto 0);
		signal wua_state 								: out WRITE_UNALIGNED_STATE;
		signal wua_state_after_clear_start 		: inout WRITE_UNALIGNED_STATE;
		signal write_unaligned_final 				: out std_logic;
		signal mcu_ll_address						: out std_logic_vector(31 downto 2);
		signal mcu_ll_xfer_write					: out	std_logic; --if high we are going to write to the mcu_ll
		signal mcu_ll_xfer_count					: out std_logic_vector(11 downto 0);
		signal mcu_ll_start_strobe_sig 			: out std_logic ) is
	begin
		hl_state <= MCU_HL_WRITE_UNALIGNED_XFER;
		mu_buss_write_sig <= '1'; --only mu buss writes from this point on
		write_unaligned_final <= '0';
		mu_next_address_xfer_end_value <= xfer_num_transfers -1;
		
		if xfer_num_transfers = X"001" then
			--start the transfer of the aligned data from the dram
			SetupUnAlignedHeadXFER(mcu_ll_address, mcu_ll_xfer_write, mcu_ll_xfer_count, mcu_ll_start_strobe_sig, mu_buss_write_sig);
			--go ahead and get the first data transfer ready from the MU
			StartMUBodyXFER(mu_buss_write_sig, mu_buss_strobe_sig, mu_buss_address, mu_next_address);
			wua_state <= WRITE_UNALIGNED_CLEAR_LL_START_STROBE;
			wua_state_after_clear_start <= WRITE_UNALIGNED_READ_SINGLE;
		else
			if xfer_first_aligned = '0' then
				wua_state <= WRITE_UNALIGNED_CLEAR_LL_START_STROBE;
				wua_state_after_clear_start <= WRITE_UNALIGNED_READ_HEAD;
				SetupUnAlignedHeadXFER(mcu_ll_address, mcu_ll_xfer_write, mcu_ll_xfer_count, mcu_ll_start_strobe_sig, mu_buss_write_sig);
			else --if xfer_final_aligned = '0' then
				SetupUnAlignedTailXFER(wua_state, wua_state_after_clear_start, mcu_ll_address, mcu_ll_xfer_write, mcu_ll_xfer_count, mcu_ll_start_strobe_sig);
			end if; --else --err.. front and end aligned... SetupUnAlignedBodyXFER();
		end if;
	end StartWriteUnAligned;

	procedure OutputUnAlignedData(
		signal xfer_alignment		: in std_logic_vector(1 downto 0);
		signal aligned_dram_data 	: inout std_logic_vector(31 downto 0);
		signal mu_buss_data_in 		: in std_logic_vector(31 downto 0);
		signal mcu_ll_data_out 		: out std_logic_vector(31 downto 0) ) is
	begin
		case xfer_alignment is
		--when "00" => --unaligned writes occur if _either_ the front or end is unaligned, so this must be implemented if the front is aligned
		--	mcu_ll_data_out <= mu_buss_data_in;
		when "01" =>
			mcu_ll_data_out <= aligned_dram_data(31 downto 24) & mu_buss_data_in(31 downto 8);
			aligned_dram_data <= mu_buss_data_in(7 downto 0) & X"000000";
		when "10" =>
			mcu_ll_data_out <= aligned_dram_data(31 downto 16) & mu_buss_data_in(31 downto 16);
			aligned_dram_data <= mu_buss_data_in(15 downto 0) & X"0000";
		when others => -- "11" =>
			mcu_ll_data_out <= aligned_dram_data(31 downto 8) & mu_buss_data_in(31 downto 24);
			aligned_dram_data <= mu_buss_data_in(23 downto 0) & X"00";
		end case;
	end OutputUnAlignedData;

	procedure OutputFinalUnAlignedData(
		signal xfer_alignment				: in std_logic_vector(1 downto 0);
		signal xfer_tail_address			: in std_logic_vector(31 downto 0);
		signal aligned_dram_data 			: in std_logic_vector(31 downto 0); 
		signal mu_buss_data_in	 			: in std_logic_vector(31 downto 0);
		signal mcu_ll_data_out 				: out std_logic_vector(31 downto 0);
		signal aligned_write_tail_data 	: in std_logic_vector(31 downto 0) ) is
		variable dataToWrite : std_logic_vector(31 downto 0);
	begin
		case xfer_alignment is
		when "00" =>
			dataToWrite := mu_buss_data_in;
		when "01" =>
			dataToWrite := aligned_dram_data(31 downto 24) & mu_buss_data_in(31 downto 8);
		when "10" =>
			dataToWrite := aligned_dram_data(31 downto 16) & mu_buss_data_in(31 downto 16);
		when others =>
			dataToWrite := aligned_dram_data(31 downto 8) & mu_buss_data_in(31 downto 24);
		end case;
		
		case xfer_tail_address(1 downto 0) is --should never be 0
		when "01" =>
			mcu_ll_data_out <= dataToWrite(31 downto 24) & aligned_write_tail_data(23 downto 0);
		when "10" =>
			mcu_ll_data_out <= dataToWrite(31 downto 16) & aligned_write_tail_data(15 downto 0);
		when others => --"11"
			mcu_ll_data_out <= dataToWrite(31 downto 8) & aligned_write_tail_data(7 downto 0);
		end case;
	end OutputFinalUnAlignedData;
										
	procedure HandleWriteUnAligned(
		signal mu_buss_address						: out std_logic_vector(11 downto 0);
		signal mu_buss_data_in 						: in std_logic_vector(31 downto 0);
		signal mu_buss_write_sig 					: out std_logic;
		signal mu_buss_strobe_sig 					: out std_logic;
		signal mu_next_address 						: inout std_logic_vector(11 downto 0);
		signal mu_next_address_xfer_end_value 	: in std_logic_vector(11 downto 0);
		signal xfer_alignment 						: in std_logic_vector(1 downto 0);	
		signal hl_state 								: out MCU_HL_STATE;
		signal aligned_dram_data 					: inout std_logic_vector(31 downto 0);
		signal wua_state 								: inout WRITE_UNALIGNED_STATE;
		signal wua_state_after_clear_start 		: inout WRITE_UNALIGNED_STATE;
		signal write_unaligned_final				: inout std_logic;
		signal aligned_write_tail_data 			: inout std_logic_vector(31 downto 0);
		signal mcu_ll_data_out 						: out std_logic_vector(31 downto 0);
		signal mcu_ll_data_in						: in std_logic_vector(31 downto 0);
		signal mcu_ll_address						: out std_logic_vector(31 downto 2);
		signal mcu_ll_xfer_write					: out	std_logic; --if high we are going to write to the mcu_ll
		signal mcu_ll_xfer_count					: out std_logic_vector(11 downto 0);
		signal mcu_ll_start_strobe_sig			: inout std_logic ) is
	begin
		case wua_state is
		when WRITE_UNALIGNED_CLEAR_LL_START_STROBE =>
			mcu_ll_start_strobe_sig <= '0';
			wua_state <= wua_state_after_clear_start;
		when WRITE_UNALIGNED_READ_SINGLE =>
			if mcu_ll_data_valid_done = '1' then
				aligned_dram_data <= mcu_ll_data_in;
				wua_state <= WRITE_UNALIGNED_READ_SINGLE_WAIT_LL_IDLE;
			end if;
		when WRITE_UNALIGNED_READ_SINGLE_WAIT_LL_IDLE =>
			if mcu_ll_idle = '1' then
				wua_state <= WRITE_UNALIGNED_CLEAR_LL_START_STROBE;
				wua_state_after_clear_start <= WRITE_UNALIGNED_SINGLE_COMPLETING;
				mcu_ll_xfer_write	<= '1';
				mcu_ll_xfer_count <= X"001";
				mcu_ll_start_strobe_sig <= '1';					
				mu_buss_strobe_sig <= '0';
				--handle a write that only involves a single buss aligned location. original data bytes
				--may remain at the start and or at the end of the byte(s) that are being written.
				case ( xfer_alignment & req_num_bytes(1 downto 0) ) is
				when ( "00" & "01" ) =>
					mcu_ll_data_out <= mu_buss_data_in(31 downto 24) & aligned_dram_data(23 downto 0);
				when ( "00" & "10" ) =>
					mcu_ll_data_out <= mu_buss_data_in(31 downto 16) & aligned_dram_data(15 downto 0);
				when ( "00" & "11" ) =>
					mcu_ll_data_out <= mu_buss_data_in(31 downto 8) & aligned_dram_data(7 downto 0);
					--hack!? would be very easy to do a mcu_data_out <= mu_data_buss here.
				when ( "01" & "01" ) =>
					mcu_ll_data_out <= aligned_dram_data(31 downto 24) & mu_buss_data_in(31 downto 24) & aligned_dram_data(15 downto 0);
				when ( "01" & "10" ) =>
					mcu_ll_data_out <= aligned_dram_data(31 downto 24) & mu_buss_data_in(31 downto 16) & aligned_dram_data(7 downto 0);
				when ( "01" & "11" ) =>
					mcu_ll_data_out <= aligned_dram_data(31 downto 24) & mu_buss_data_in(31 downto 8);
				when ( "10" & "01" ) =>
					mcu_ll_data_out <= aligned_dram_data(31 downto 16) & mu_buss_data_in(31 downto 24) & aligned_dram_data(7 downto 0);
				when ( "10" & "10" ) =>
					mcu_ll_data_out <= aligned_dram_data(31 downto 16) & mu_buss_data_in(31 downto 16);
				when ( "11" & "01" ) =>
					mcu_ll_data_out <= aligned_dram_data(31 downto 8) & mu_buss_data_in(31 downto 24); --hack. unchecked change for size fix
				when others =>
					null; -- a bug.
				end case;
			end if;
		when WRITE_UNALIGNED_SINGLE_COMPLETING =>
			if mcu_ll_idle = '1' then --mcu_ll_data_valid_done = '1'
				hl_state <= MCU_HL_IDLE;
			end if;
		when WRITE_UNALIGNED_READ_HEAD =>
			if	mcu_ll_data_valid_done = '1' then
				aligned_dram_data <= mcu_ll_data_in;
				wua_state <= WRITE_UNALIGNED_READ_HEAD_WAIT_LL_IDLE;
			end if;
		when WRITE_UNALIGNED_READ_HEAD_WAIT_LL_IDLE =>
			if mcu_ll_idle = '1' then
				if xfer_final_aligned = '0' then
					SetupUnAlignedTailXFER(wua_state, wua_state_after_clear_start, mcu_ll_address, mcu_ll_xfer_write, mcu_ll_xfer_count, mcu_ll_start_strobe_sig);
				else
					wua_state <= WRITE_UNALIGNED_BODY_STARTUP;
					StartMUBodyXFER(mu_buss_write_sig, mu_buss_strobe_sig, mu_buss_address, mu_next_address);
				end if;
			end if;
		when WRITE_UNALIGNED_READ_TAIL =>
			if	mcu_ll_data_valid_done = '1' then
				aligned_write_tail_data <= mcu_ll_data_in;
				wua_state <= WRITE_UNALIGNED_READ_TAIL_WAIT_LL_IDLE;
			end if;
		when WRITE_UNALIGNED_READ_TAIL_WAIT_LL_IDLE =>
			if mcu_ll_idle = '1' then
				wua_state <= WRITE_UNALIGNED_BODY_STARTUP_DELAY;
				StartMUBodyXFER(mu_buss_write_sig, mu_buss_strobe_sig, mu_buss_address, mu_next_address);
			end if;
		when WRITE_UNALIGNED_BODY_STARTUP_DELAY => --we must wait one clock cycle for the MU to present valid data.
			wua_state <= WRITE_UNALIGNED_BODY_STARTUP;
		when WRITE_UNALIGNED_BODY_STARTUP => --the mu data should be ready now.	
			OutputUnAlignedData(xfer_alignment, aligned_dram_data, mu_buss_data_in, mcu_ll_data_out);
			mcu_ll_address <= req_byte_address(31 downto 2);
			mcu_ll_xfer_count <= xfer_num_transfers;
			mcu_ll_xfer_write <= '1';
			mcu_ll_start_strobe_sig <= '1';
			mu_buss_strobe_sig <= '0';--clock data out (poly-scanline hack)

			wua_state <= WRITE_UNALIGNED_CLEAR_LL_START_STROBE;
			wua_state_after_clear_start <= WRITE_UNALIGNED_BODY;
			
			mu_buss_address <= mu_next_address;
			if mu_next_address = mu_next_address_xfer_end_value then
				write_unaligned_final <= '1';
			else
				mu_next_address <= mu_next_address + 1;
			end if;
		when WRITE_UNALIGNED_BODY =>
			if	mcu_ll_data_valid_done = '1' then
				if write_unaligned_final = '1' then
					OutputFinalUnalignedData(xfer_alignment, xfer_tail_address, aligned_dram_data, mu_buss_data_in, mcu_ll_data_out, aligned_write_tail_data);
					wua_state <= WRITE_UNALIGNED_COMPLETING;
				else
					OutputUnAlignedData(xfer_alignment, aligned_dram_data, mu_buss_data_in, mcu_ll_data_out);
					mu_buss_address <= mu_next_address;
					if mu_next_address = mu_next_address_xfer_end_value then --this is the last transfer
						write_unaligned_final <= '1';
					else
						mu_next_address <= mu_next_address + 1;
					end if;
				end if;
				--output the current data value
				mu_buss_strobe_sig <= '0';--clock data out (poly-scanline hack)
			else
				mu_buss_strobe_sig <= '1';--clock data out (poly-scanline hack)
			end if;
		when WRITE_UNALIGNED_COMPLETING =>
			if mcu_ll_idle = '1' then
				mu_buss_strobe_sig <= '0';
				hl_state <= MCU_HL_IDLE;
			end if;
		when others => null;
		end case;
	end HandleWriteUnAligned;

	--HandleIdle already set up the active mu to transfer the configuration data.
	procedure HandleStartup(
							signal hl_state						: inout MCU_HL_STATE;
							signal hl_after_waitstate_state 	: out MCU_HL_STATE;
							signal ra_state						: inout READ_ALIGNED_STATE;
							signal wa_state						: inout WRITE_ALIGNED_STATE;
							signal rua_state 						: out READ_UNALIGNED_STATE;
							signal wua_state 						: out WRITE_UNALIGNED_STATE;
							signal wua_state_after_clear_start : inout WRITE_UNALIGNED_STATE;
							signal write_unaligned_final		: inout std_logic;
							signal mu_next_address				: out std_logic_vector(11 downto 0);
							signal mu_next_address_xfer_end_value 	: inout std_logic_vector(11 downto 0);
							signal mu_buss_address				: out std_logic_vector(11 downto 0);
							signal mu_buss_data_in				: in std_logic_vector(31 downto 0);
							signal mu_buss_data_out				: out std_logic_vector(31 downto 0);
							signal mu_buss_write_sig			: out std_logic;
							signal mu_buss_strobe_sig			: out std_logic;
							signal mcu_ll_address				: out std_logic_vector(31 downto 2);
							signal mcu_ll_xfer_write			: out	std_logic;
							signal mcu_ll_xfer_count			: out std_logic_vector(11 downto 0);
							signal mcu_ll_start_strobe_sig	: out std_logic;
							signal req_byte_address				: inout std_logic_vector(31 downto 0);
							signal req_num_bytes					: inout std_logic_vector(23 downto 0);
							signal req_write						: inout std_logic;
							signal xfer_tail_address			: inout std_logic_vector(31 downto 0);
							signal xfer_num_transfers 			: inout std_logic_vector(11 downto 0);
							signal xfer_first_aligned 			: inout std_logic;
							signal xfer_final_aligned 			: inout std_logic;
							signal xfer_alignment 				: inout std_logic_vector(1 downto 0) ) is
	begin
		case hl_state is
		when MCU_HL_STARTUP=>
			--read current data
			req_write <= mu_buss_data_in(31);
			req_num_bytes <= mu_buss_data_in(23 downto 0);
			mu_buss_address <= X"FFD";
			--the strobe should still be high from the scheduler
			hl_after_waitstate_state <= MCU_HL_STARTUP_2;
			hl_state <= MCU_HL_WAITSTATE; --it will take two clockcycles before the mu_buss_changes
		when MCU_HL_STARTUP_2=>
			req_byte_address <= mu_buss_data_in;
			xfer_alignment <= mu_buss_data_in(1 downto 0);
			xfer_tail_address <= mu_buss_data_in + req_num_bytes;
			mu_buss_strobe_sig <= '0'; --we need a clock cycle to process this.
			hl_state <= MCU_HL_STARTUP_VALIDATE;
		when MCU_HL_STARTUP_VALIDATE=>
			if xfer_tail_address <= req_byte_address then --there was an error. either the number of memory addesses was zero, or there was an overflow
				hl_after_waitstate_state <= MCU_HL_IDLE; --HACK. with the process handling sched_busy this does nothing
				hl_state <= MCU_HL_WAITSTATE; --we need the mcu scheduler to deassert its irq before we are actually idle.
			else
				--get ready to do much more work
				case xfer_alignment & req_num_bytes(1 downto 0) is
				when "00" & "00" =>
					xfer_num_transfers <= "00" & req_num_bytes(11 downto 2);
				when "01" & "00" | "10" & "00" | "11" & "00"  =>
					xfer_num_transfers <= ( "00" & req_num_bytes(11 downto 2) ) + 1;
				when  "00" & "01" | "01" & "01" | "10" & "01" | "11" & "01" =>
					xfer_num_transfers <= ( "00" & req_num_bytes(11 downto 2) ) + 1;
				when "00" & "10" | "01" & "10" | "10" & "10" =>
					xfer_num_transfers <= ( "00" & req_num_bytes(11 downto 2) ) + 1;				
				when "00" & "11" | "01" & "11" =>
					xfer_num_transfers <= ( "00" & req_num_bytes(11 downto 2) ) + 1;
				when others =>
				--when "10" & "11" =>
				--when "11" & "10" =>
				--when "11" & "11" =>
					xfer_num_transfers <= ( "00" & req_num_bytes(11 downto 2) ) + 2;
				end case;
	
				if xfer_alignment = "00" then
					xfer_first_aligned <= '1';
				else
					xfer_first_aligned <= '0';
				end if;
				
				if xfer_tail_address(1 downto 0) = "00" then
					xfer_final_aligned <= '1';
				else
					xfer_final_aligned <= '0';
				end if;					
				hl_state <= MCU_HL_STARTUP_CONFIGURE;
			end if;
		when 	MCU_HL_STARTUP_CONFIGURE =>
			if req_write = '1' then
				if xfer_first_aligned = '1' and xfer_final_aligned = '1' then
					StartWriteAligned(hl_state, wa_state, mu_buss_address, mu_buss_write_sig, mu_buss_strobe_sig, mu_next_address, mcu_ll_address, mcu_ll_xfer_write, mcu_ll_xfer_count, mcu_ll_start_strobe_sig);
				else
					StartWriteUnAligned(xfer_num_transfers, xfer_first_aligned, mu_buss_write_sig, mu_buss_strobe_sig, mu_buss_address, hl_state,  mu_next_address, mu_next_address_xfer_end_value, wua_state, wua_state_after_clear_start, write_unaligned_final,
												mcu_ll_address, mcu_ll_xfer_write, mcu_ll_xfer_count, mcu_ll_start_strobe_sig);
				end if;
			else --a read
				if xfer_first_aligned = '1' then -- and xfer_final_aligned = '1' then
					StartReadAligned(hl_state, ra_state, mu_buss_write_sig, mu_next_address, mcu_ll_address, mcu_ll_xfer_write, mcu_ll_xfer_count, mcu_ll_start_strobe_sig);
				else
					StartReadUnAligned(hl_state, rua_state, mu_buss_write_sig, mu_next_address, mcu_ll_address, mcu_ll_xfer_write, mcu_ll_xfer_count, mcu_ll_start_strobe_sig);
				end if;
			end if;
		when others => --a bug
			hl_state <= MCU_HL_IDLE;
		end case;
	end HandleStartup;

begin

	mu_buss_write <= mu_buss_write_sig;
	mu_buss_strobe <= mu_buss_strobe_sig;
	mu_buss_data_in <= mu_buss_data;

	process(mcu_ll_start_strobe_sig)
	begin
			mcu_ll_start_strobe <= mcu_ll_start_strobe_sig;
	end process;

	process(mu_buss_data_out, mu_buss_write_sig)
	begin
		mu_buss_write <= mu_buss_write_sig;
	  if mu_buss_write_sig = '1' then
			mu_buss_data <= mu_buss_data_out;
		else
			mu_buss_data <= (others => 'Z');
	end if;
	end process;

	process(hl_state)
	begin
		if hl_state = MCU_HL_IDLE then
			mcu_hl_busy <= '0';
		else
			mcu_hl_busy <= '1';
		end if;
	end process;

	process(tvc_clock)
	begin
	if tvc_clock = '1' and tvc_clock'event then
	if tvc_reset = '1' then
		hl_state <= MCU_HL_IDLE;
		hl_after_waitstate_state <= MCU_HL_IDLE;

		req_byte_address <= (others => '0');
		req_num_bytes <= (others => '0');
		req_write<= '1';
		
		xfer_tail_address <= (others => '0');
		xfer_num_transfers <= (others => '0');
		xfer_first_aligned <= '1';
		xfer_final_aligned <= '1';
		xfer_alignment <= "00";
		aligned_dram_data <= X"0000_0000";
		aligned_write_tail_data <= X"0000_0000";
		
		mu_buss_address <= X"FFC"; --we need to keep the mu data buss configured to input the first mu configuration register.
		mu_buss_write_sig <= '0';
		mu_buss_strobe_sig <= '1';
		mu_buss_data_out <= (others => '0');
		mu_next_address <= (others => '0');

		write_unaligned_final <= '0';
		--mcu_hl_sched_busy	<= '0'; --handled by its own process

		mcu_ll_data_out <= (others => '0');
		mcu_ll_address <= (others => '0');
		mcu_ll_xfer_write	<= '0';
		mcu_ll_xfer_count <= (others => '0');
		mcu_ll_start_strobe_sig <= '0';
	else --not a reset
		case hl_state is
		when MCU_HL_IDLE =>
			--ugly to have this set this every time.
			mu_buss_address <= X"FFC";
			mu_buss_write_sig <= '0';
			mu_buss_strobe_sig <= '1';
			if mcu_hl_start_strobe = '1' then
				hl_state <= MCU_HL_STARTUP;
			end if;
		when MCU_HL_STARTUP | MCU_HL_STARTUP_2 | MCU_HL_STARTUP_VALIDATE | MCU_HL_STARTUP_CONFIGURE =>
			HandleStartup(hl_state, hl_after_waitstate_state, ra_state, wa_state, rua_state, wua_state, wua_state_after_clear_start, write_unaligned_final, mu_next_address, mu_next_address_xfer_end_value,
							mu_buss_address, mu_buss_data_in, mu_buss_data_out, mu_buss_write_sig, mu_buss_strobe_sig,
							mcu_ll_address, mcu_ll_xfer_write, mcu_ll_xfer_count, mcu_ll_start_strobe_sig,
							req_byte_address, req_num_bytes, req_write,
							xfer_tail_address, xfer_num_transfers, xfer_first_aligned, xfer_final_aligned, xfer_alignment);
		when MCU_HL_READ_ALIGNED_XFER =>
			HandleReadAligned(hl_state, ra_state, mu_buss_address, mu_buss_data_out, mu_buss_write_sig, mu_buss_strobe_sig, mu_next_address,
									mcu_ll_data_in, mcu_ll_data_valid_done, mcu_ll_start_strobe_sig, mcu_ll_idle);
		when MCU_HL_READ_UNALIGNED_XFER =>
			HandleReadUnAligned(hl_state, rua_state, xfer_num_transfers, aligned_dram_data, mu_buss_address, mu_buss_data_out, mu_buss_write_sig, mu_buss_strobe_sig, mu_next_address, mu_next_address_xfer_end_value,
									mcu_ll_data_in, mcu_ll_data_valid_done, mcu_ll_start_strobe_sig, mcu_ll_idle);
		when MCU_HL_WRITE_ALIGNED_XFER =>
			HandleWriteAligned(hl_state, wa_state,
								mu_buss_address, mu_buss_data_in, mu_buss_write_sig, mu_buss_strobe_sig, mu_next_address,
								mcu_ll_data_out, mcu_ll_data_valid_done, mcu_ll_start_strobe_sig, mcu_ll_idle);
		when MCU_HL_WRITE_UNALIGNED_XFER =>
			HandleWriteUnAligned(mu_buss_address, mu_buss_data_in, mu_buss_write_sig, mu_buss_strobe_sig, mu_next_address, mu_next_address_xfer_end_value, xfer_alignment,
										hl_state, aligned_dram_data, wua_state, wua_state_after_clear_start, write_unaligned_final, aligned_write_tail_data,
										mcu_ll_data_out, mcu_ll_data_in, mcu_ll_address, mcu_ll_xfer_write, mcu_ll_xfer_count, mcu_ll_start_strobe_sig);
		when 	MCU_HL_WAITSTATE =>
			hl_state <= hl_after_waitstate_state;
		end case;
	end if; --if reset
	end if; --clock event
	end process;
end Behavioral;

