 library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
library UNISIM;
use UNISIM.VComponents.all;

entity scanline_cache is
	Port(
	--memory user access bus
	mu_bus_address 		: in std_logic_vector(11 downto 0);
	mu_bus_data			: inout std_logic_vector(31 downto 0);
	mu_bus_write			: in  std_logic;
	mu_bus_strobe			: in  std_logic;

	sc_irq					: out std_logic;
	sc_select				: in std_logic;

	tvc_clock 			: in 	std_logic;
	tvc_reset 			: in 	std_logic;

	sc_start_fill			: in		std_logic; --change the state to ask the mcu with data starting from the address in sc_start_fill_addr
	sc_start_fill_addr	: in		std_logic_vector(23 downto 0); 

	--these signals only connect to the second port of the sram and operate in the pixel clock domain
	--they are mostly ignored by the logic in this class.
	pixel_clock			: in std_logic;
	next_pix_addr		: in std_logic_vector(11 downto 0);
	next_pix_values	: out std_logic_vector(31 downto 0)-- when in drain mode this is 'updated' on every clock.
	);
end scanline_cache;

architecture Behavioral of scanline_cache is

	type SC_CURR_STATE is (SC_IDLE_STATE, SC_MCU_DRIVEN); --, SC_PIXEL_DRAIN);
	signal curr_state : SC_CURR_STATE;
	 
	component RAMB4_S16_S16
	port (WEA, ENA, RSTA, CLKA, WEB, ENB, RSTB, CLKB: in STD_LOGIC;
	ADDRA: in STD_LOGIC_VECTOR(7 downto 0);
	DIA: in STD_LOGIC_VECTOR(15 downto 0);
	--DOA: out STD_LOGIC_VECTOR(15 downto 0);
	ADDRB: in STD_LOGIC_VECTOR(7 downto 0);
	--DIB: in STD_LOGIC_VECTOR(15 downto 0);
	DOB: out STD_LOGIC_VECTOR(15 downto 0));
	end component;
	
	--signals for the block ram
	signal brWE 	: std_logic;
	signal brEN  	: std_logic;
	signal brRST	: std_logic;

	signal brDI	: std_logic_vector(31 downto 0);
	--signal brDO	: std_logic_vector(31 downto 0);

	signal brADDR		: std_logic_vector(7 downto 0);

	signal scanline_addr_start 	: std_logic_vector(23 downto 0); --a local copy of the data from sc_start_fill_addr

 	signal mu_bus_data_in : std_logic_vector(31 downto 0);
	signal mu_bus_data_out : std_logic_vector(31 downto 0);

	signal mcu_selected : std_logic;

	begin
	high_word : RAMB4_S16_S16
		port map(WEA=>brWE, ENA=>brEN, RSTA=>brRST, CLKA=>tvc_clock,
					ADDRA=>brADDR, DIA=>brDI(31 downto 16), --DOA=>brDO(31 downto 16)
					WEB=>'0', ENB=>'1', RSTB=>'0', CLKB=>pixel_clock,
					ADDRB=>next_pix_addr(9 downto 2), DOB=>next_pix_values(31 downto 16) );
	low_word	: RAMB4_S16_S16
 		port map(WEA=>brWE, ENA=>brEN, RSTA=>brRST, CLKA=>tvc_clock,
					ADDRA=>brADDR, DIA=>brDI(15 downto 0), --DOA=>brDO(15 downto 0) );				
					WEB=>'0', ENB=>'1', RSTB=>'0', CLKB=>pixel_clock,
					ADDRB=>next_pix_addr(9 downto 2), DOB=>next_pix_values(15 downto 0) );

	mu_bus_data_in <= mu_bus_data;
	process(mu_bus_data_out, sc_select, mu_bus_strobe, mu_bus_write)
	begin
		if sc_select = '1' and mu_bus_strobe = '1' and mu_bus_write = '0' then
			mu_bus_data <= mu_bus_data_out;
		else
			mu_bus_data <= (others => 'Z');
		end if;
	end process;

 	process(tvc_clock)
	begin
	if tvc_clock = '1' and tvc_clock'event then
	if tvc_reset = '1' then
		curr_state <= SC_IDLE_STATE;
		brRST <= '0';
		brEN <= '1';
		brWE <= '0';
		--brDI <= X"0000_0000";
		sc_irq <= '0';
		mcu_selected <= '0';
	else
		brRST <= '0';
		case curr_state is
		when SC_IDLE_STATE =>
			if sc_start_fill = '1' then
				scanline_addr_start <= sc_start_fill_addr;
				curr_state <= SC_MCU_DRIVEN;
				mcu_selected <= '0';
				sc_irq <= '1';
				brEN <= '1';
			end if;
		when SC_MCU_DRIVEN =>
			if sc_select = '0' and mcu_selected = '1' then
				curr_state <= SC_IDLE_STATE;
				sc_irq <= '0';
				brWE <= '0'; --necessary?
				--brEN <= '0';
			elsif sc_select = '1' then
				mcu_selected <= '1';
				if mu_bus_strobe = '1' then
					if mu_bus_write = '1' then
						brADDR <= mu_bus_address(7 downto 0);
						--brDI <= X"c0380700";
						--brDI <= mu_bus_address(7 downto 0) & mu_bus_address(7 downto 0) & mu_bus_address(7 downto 0) & mu_bus_address(7 downto 0);
						brDI <= mu_bus_data_in;
						brWE <= '1';
					else
						brWE <= '0';
						if mu_bus_address = X"FFC" then
							mu_bus_data_out <=  X"00" & X"000280";
						elsif mu_bus_address = X"FFD" then
							mu_bus_data_out <= X"00" & scanline_addr_start;
						else
							mu_bus_data_out <= X"0000_0000"; --a problem						
						end if;
					end if; --bus_write
				else --bus_strobe
					brWE <= '0';
				end if;
	end if;
		end case;
	end if;
	end if;
	end process;
end Behavioral;
