library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
library UNISIM;
use UNISIM.VComponents.all;

--library work;
--use work.all;

entity tvc is		    
    Port (		 
		tvc_clock 				: in 	std_logic;
		tvc_reset_in 			: in 	std_logic;

		led_out					: out	std_logic;

		epp_data 				: inout	std_logic_vector(7 downto 0);
		epp_write 				: in 	std_logic;
		epp_wait 				: out 	std_logic;
		epp_data_strobe 		: in 	std_logic;
		epp_address_strobe 	: in 	std_logic;
 		epp_irq					: out	std_logic;
 --         epp_reset 		: in 	std_logic;

		disp1_pix_value 		: out 	std_logic_vector(7 downto 0);
		disp1_horiz_sync 		: out 	std_logic;
		disp1_vert_sync 		: out 	std_logic;

		dram_d_buss				: inout std_logic_vector(31 downto 0);
		dram_a_buss				: out std_logic_vector(11 downto 0);
		dram_ras_0				: out std_logic;
		dram_ras_1				: out std_logic;
		dram_ras_2				: out std_logic;
		dram_ras_3				: out std_logic;
		dram_cas_0				: out std_logic;
		dram_cas_1				: out std_logic;
		dram_cas_2				: out std_logic;
		dram_cas_3				: out std_logic;
		dram_we					: out std_logic
	  );
end tvc;

architecture Behavioral of tvc is
	component reset_sync
	Port (
	sig_in	: in	STD_LOGIC;
	sig_out	: out	STD_LOGIC;
	clock		: in	STD_LOGIC
	);
	end component;

	component pix_pusher
	Port(
	mu_buss_address 	: in std_logic_vector(11 downto 0);
	mu_buss_data		: inout std_logic_vector(31 downto 0);
	mu_buss_write		: in std_logic;
	mu_buss_strobe		: in std_logic;
	odd_sc_irq			: out std_logic;
	odd_sc_select		: in std_logic;
	even_sc_irq			: out std_logic;
	even_sc_select		: in std_logic;
	tvc_clock			: in 	std_logic;
	tvc_reset			: in 	std_logic;
	disp_pix_value 	: out 	std_logic_vector(7 downto 0);
	disp_horiz_sync 	: out 	std_logic;
	disp_vert_sync 	: out 	std_logic
	);
	end component;

	component eppcu
	Port(
	epp_data					: inout std_logic_vector(7 downto 0);
	epp_write				: in 	std_logic;
	epp_wait					: out std_logic;
	epp_data_strobe 		: in 	std_logic;
	epp_address_strobe 	: in 	std_logic;
	epp_irq					: out	std_logic;
	tvc_clock				: in std_logic;
	tvc_reset				: in std_logic;
	br_fb_address 			: out std_logic_vector(31 downto 0);
	br_fb_data				: in std_logic_vector(31 downto 0);
	br_start_strobe		: out std_logic;
	br_idle					: in std_logic;
	bw_fb_address 			: out std_logic_vector(31 downto 0);
	bw_fb_data				: out std_logic_vector(31 downto 0);
	bw_start_strobe		: out std_logic;
	bw_idle					: in std_logic;
	bs_fb_address 			: out std_logic_vector(31 downto 2);
	bs_num_bytes			: out std_logic_vector(23 downto 2);
	bs_fb_value				: out std_logic_vector(31 downto 0);
	bs_start_strobe		: out std_logic;
	bs_idle					: in std_logic;
--	rpw_fb_address 		: out std_logic_vector(31 downto 0);
--	rpw_pix_data			: out std_logic_vector(7 downto 0);
--	rpw_start_strobe		: out std_logic;
--	rpw_idle					: in std_logic;
	ps_fb_address 			: out std_logic_vector(31 downto 0);
	ps_db_address 			: out std_logic_vector(31 downto 0);
	ps_start_z_depth		: out std_logic_vector(31 downto 0);
	ps_z_increment			: out std_logic_vector(31 downto 0);
	ps_scanline_length	: out std_logic_vector(15 downto 0);
	ps_pix_value			: out std_logic_vector(7 downto 0);
	ps_start_strobe		: out std_logic;
	ps_idle					: in std_logic
	);
	end component;

	component mcu_struct is
    Port (
		dram_d_buss		: inout std_logic_vector(31 downto 0);
		dram_a_buss 	: out std_logic_vector(11 downto 0);
		dram_ras_0 		: out std_logic;
		dram_ras_1 		: out std_logic;
		dram_ras_2 		: out std_logic;
		dram_ras_3 		: out std_logic;
		dram_cas_0 		: out std_logic;
		dram_cas_1 		: out std_logic;
		dram_cas_2 		: out std_logic;
		dram_cas_3 		: out std_logic;
		dram_we			: out std_logic;
		tvc_clock		: in	std_logic;
		tvc_reset		: in 	std_logic;
		mu_buss_address 	: out std_logic_vector(11 downto 0);
		mu_buss_data		: inout std_logic_vector(31 downto 0);
		mu_buss_write		: out  std_logic;
		mu_buss_strobe			: out  std_logic;
		odd_sc_irq				: in std_logic;
		odd_sc_select			: out std_logic;
		even_sc_irq				: in std_logic;
		even_sc_select			: out std_logic;
		block_reader_irq		: in std_logic;
		block_reader_select	: out std_logic;
		block_writer_irq		: in std_logic;
		block_writer_select	: out std_logic;				 
		rp_writer_irq			: in std_logic;
		rp_writer_select		: out std_logic;
		line_drawer_irq		: in std_logic;
		line_drawer_select	: out std_logic;
		poly_scanline_irq		: in std_logic;
		poly_scanline_select : out std_logic
		);	
	end component;
 
	component block_reader is
	Port(
	mu_buss_address 	: in std_logic_vector(11 downto 0);
	mu_buss_data		: inout std_logic_vector(31 downto 0);
	mu_buss_write		: in std_logic; --if high the MCU is going to write to a mu
	mu_buss_strobe			: in std_logic;
	block_reader_irq		: out std_logic;
	block_reader_select	: in std_logic;
	tvc_clock			: in	std_logic;
	tvc_reset			: in std_logic;
	epp_fb_address 	: in std_logic_vector(31 downto 0);
	epp_fb_data		: out std_logic_vector(31 downto 0);
	epp_start_strobe	: in std_logic;
	epp_idle			: out std_logic
	);
	end component;

	component rp_writer is
	Port(
		mu_buss_address 	: in std_logic_vector(11 downto 0);
		mu_buss_data		: inout std_logic_vector(31 downto 0);
		mu_buss_write		: in std_logic;
		mu_buss_strobe		: in std_logic;
		rp_writer_irq		: out std_logic;
		rp_writer_select	: in std_logic;
		tvc_clock			: in	std_logic;
		tvc_reset			: in std_logic;
		epp_fb_address 	: in std_logic_vector(31 downto 0);
		epp_pix_data		: in std_logic_vector(7 downto 0);
		epp_start_strobe	: in std_logic;
		epp_idle			: out std_logic
	  );
	end component;

	component block_writer is
	Port(
	mu_buss_address 	: in std_logic_vector(11 downto 0);
	mu_buss_data		: inout std_logic_vector(31 downto 0);
	mu_buss_write		: in  std_logic;
	mu_buss_strobe		: in  std_logic;
	block_writer_irq		: out std_logic;
	block_writer_select	: in std_logic;
	tvc_clock			: in		std_logic;
	tvc_reset			: in 	std_logic;
	epp_fb_address 	: in std_logic_vector(31 downto 0);
	epp_fb_data		: in std_logic_vector(31 downto 0);
	epp_start_strobe	: in std_logic;
	epp_idle			: out std_logic
	);
	end component;

	component block_set is
	Port(
	mu_buss_address 	: in std_logic_vector(11 downto 0);
	mu_buss_data		: inout std_logic_vector(31 downto 0);
	mu_buss_write		: in  std_logic;
	mu_buss_strobe		: in  std_logic;
	block_set_irq		: out std_logic;
	block_set_select	: in std_logic;
	tvc_clock			: in std_logic;
	tvc_reset			: in std_logic;
	epp_fb_address 	: in std_logic_vector(31 downto 2);
	epp_num_bytes		: in std_logic_vector(23 downto 2);
	epp_fb_value		: in std_logic_vector(31 downto 0);
	epp_start_strobe	: in std_logic;
	epp_idle				: out std_logic
	);
	end component;

	component poly_scanline is
   Port (
	mu_buss_address 		: in std_logic_vector(11 downto 0);
	mu_buss_data			: inout std_logic_vector(31 downto 0);
	mu_buss_write			: in  std_logic;
	mu_buss_strobe			: in  std_logic;
	poly_scanline_irq		: out std_logic;
	poly_scanline_select	: in std_logic;
	tvc_clock				: in	std_logic;
	tvc_reset				: in 	std_logic;
	epp_fb_address 		: in std_logic_vector(31 downto 0);
	epp_db_address 		: in std_logic_vector(31 downto 0);
	epp_start_z_depth		: in std_logic_vector(31 downto 0);
	epp_z_increment		: in std_logic_vector(31 downto 0);
	epp_scanline_length	: in std_logic_vector(15 downto 0);
	epp_pix_value			: in std_logic_vector(7 downto 0);
	epp_start_strobe		: in std_logic;
	epp_idle					: out std_logic
	);
	end component;

--	component IBUFG
--	port (
--		 I : in STD_LOGIC;
--		 O : out STD_LOGIC);
--	end component;

	signal tvc_reset				: std_logic;
--start mcu related signals
	signal mu_buss_address		: std_logic_vector(11 downto 0);
	signal mu_buss_data			: std_logic_vector(31 downto 0);
	signal mu_buss_write			: std_logic;
	signal mu_buss_strobe		: std_logic;

	signal odd_sc_irq				: std_logic;
	signal odd_sc_select			: std_logic;
	signal even_sc_irq			: std_logic;
	signal even_sc_select		: std_logic;
	signal block_reader_irq		: std_logic;
	signal block_reader_select	: std_logic;
	signal block_writer_irq		: std_logic;
	signal block_writer_select	: std_logic;
	signal block_set_irq			: std_logic;
	signal block_set_select		: std_logic;
	signal line_drawer_irq		: std_logic;
	signal line_drawer_select	: std_logic;
	signal poly_scanline_irq	: std_logic;
	signal poly_scanline_select : std_logic;


--begin eppcu and block_reader interface logic
	signal br_epp_fb_address 	: std_logic_vector(31 downto 0);
	signal br_epp_fb_data		: std_logic_vector(31 downto 0);
	signal br_epp_start_strobe	: std_logic;
	signal br_epp_idle			: std_logic;

--begin eppcu and block_writer interface logic
	signal bw_epp_fb_address 	: std_logic_vector(31 downto 0);
	signal bw_epp_fb_data		: std_logic_vector(31 downto 0);
	signal bw_epp_start_strobe	: std_logic;
	signal bw_epp_idle			: std_logic;

--begin eppcu and block set
	signal bs_epp_fb_address 		: std_logic_vector(31 downto 2);
	signal bs_epp_num_bytes			: std_logic_vector(23 downto 2);
	signal bs_epp_fb_value			: std_logic_vector(31 downto 0);
	signal bs_epp_start_strobe		: std_logic;
	signal bs_epp_idle				: std_logic;
--begin eppcu and rp_writer
--	signal rpw_epp_fb_address 		: std_logic_vector(31 downto 0);
--	signal rpw_epp_pix_data			: std_logic_vector(7 downto 0);
--	signal rpw_epp_start_strobe	: std_logic;
--	signal rpw_epp_idle				: std_logic;

--begin eppcu and poly_scanline mu
	signal ps_epp_fb_address 		: std_logic_vector(31 downto 0);
	signal ps_epp_db_address 		: std_logic_vector(31 downto 0);
	signal ps_epp_start_z_depth	: std_logic_vector(31 downto 0);
	signal ps_epp_z_increment		: std_logic_vector(31 downto 0);
	signal ps_epp_scanline_length	: std_logic_vector(15 downto 0);
	signal ps_epp_pix_value			: std_logic_vector(7 downto 0);
	signal ps_epp_start_strobe		: std_logic;
	signal ps_epp_idle				: std_logic;


begin
  	--tvc_reset <= NOT tvc_reset_in;
	led_out <= tvc_reset;
	--led_out <= '0';
	
	line_drawer_irq <= '0';

	rsync : reset_sync
	port map( sig_in=> NOT tvc_reset_in, sig_out=>tvc_reset, clock=>tvc_clock);
	
	ppusher : pix_pusher
	port map(mu_buss_address=>mu_buss_address, mu_buss_data=>mu_buss_data, mu_buss_write=>mu_buss_write, mu_buss_strobe=>mu_buss_strobe,
				odd_sc_irq=>odd_sc_irq, odd_sc_select=>odd_sc_select, even_sc_irq=>even_sc_irq, even_sc_select=>even_sc_select,
				tvc_clock=>tvc_clock,  tvc_reset=>tvc_reset, disp_pix_value=>disp1_pix_value,
		    	disp_horiz_sync=>disp1_horiz_sync, disp_vert_sync=>disp1_vert_sync);

	control_unit : eppcu
		port map(epp_data=>epp_data, epp_write=>epp_write, epp_wait=>epp_wait, epp_data_strobe=>epp_data_strobe,
				epp_address_strobe=>epp_address_strobe, epp_irq=>epp_irq, tvc_clock=>tvc_clock, tvc_reset=>tvc_reset,
				br_fb_address=>br_epp_fb_address, br_fb_data=>br_epp_fb_data, br_start_strobe=>br_epp_start_strobe, br_idle=>br_epp_idle,
				bw_fb_address=>bw_epp_fb_address, bw_fb_data=>bw_epp_fb_data, bw_start_strobe=>bw_epp_start_strobe, bw_idle=>bw_epp_idle,
				bs_fb_address=>bs_epp_fb_address, bs_num_bytes=>bs_epp_num_bytes, bs_fb_value=>bs_epp_fb_value, bs_start_strobe=>bs_epp_start_strobe, bs_idle=>bs_epp_idle,
--				rpw_fb_address=>rpw_epp_fb_address, rpw_pix_data=>rpw_epp_pix_data, rpw_start_strobe=>rpw_epp_start_strobe, rpw_idle=>rpw_epp_idle,
				ps_fb_address=>ps_epp_fb_address, ps_db_address=>ps_epp_db_address, ps_start_z_depth=>ps_epp_start_z_depth, ps_z_increment=>ps_epp_z_increment,
				ps_scanline_length=>ps_epp_scanline_length, ps_pix_value=>ps_epp_pix_value, ps_start_strobe=>ps_epp_start_strobe, ps_idle=>ps_epp_idle);

	mem_control_unit : mcu_struct
		port map(	dram_d_buss=>dram_d_buss, dram_a_buss=>dram_a_buss,
				dram_ras_0=>dram_ras_0, dram_ras_1=>dram_ras_1, dram_ras_2=>dram_ras_2, dram_ras_3=>dram_ras_3,
				dram_cas_0=>dram_cas_0, dram_cas_1=>dram_cas_1, dram_cas_2=>dram_cas_2, dram_cas_3=>dram_cas_3,
				dram_we=>dram_we,
				tvc_clock=>tvc_clock, tvc_reset=>tvc_reset,
				mu_buss_address=>mu_buss_address, mu_buss_data=>mu_buss_data, mu_buss_write=>mu_buss_write, mu_buss_strobe=>mu_buss_strobe,
				odd_sc_irq=>odd_sc_irq, odd_sc_select=>odd_sc_select, even_sc_irq=>even_sc_irq, even_sc_select=>even_sc_select,
				block_reader_irq=>block_reader_irq, block_reader_select=>block_reader_select,
				block_writer_irq=>block_writer_irq, block_writer_select=>block_writer_select,
				rp_writer_irq=>block_set_irq, rp_writer_select=>block_set_select,
				line_drawer_irq=>line_drawer_irq, line_drawer_select=>line_drawer_select,
				poly_scanline_irq=>poly_scanline_irq, poly_scanline_select=>poly_scanline_select);

	--start memory users

	br_mu : block_reader
		port map(mu_buss_address=>mu_buss_address, mu_buss_data=>mu_buss_data, mu_buss_write=>mu_buss_write, mu_buss_strobe=>mu_buss_strobe,
				block_reader_irq=>block_reader_irq, block_reader_select=>block_reader_select,
				tvc_clock=>tvc_clock, tvc_reset=>tvc_reset,
				epp_fb_address=>br_epp_fb_address,
				epp_fb_data=>br_epp_fb_data, epp_start_strobe=>br_epp_start_strobe, epp_idle=>br_epp_idle );

	bw_mu : block_writer
		port map(mu_buss_address=>mu_buss_address, mu_buss_data=>mu_buss_data, mu_buss_write=>mu_buss_write, mu_buss_strobe=>mu_buss_strobe,
				block_writer_irq=>block_writer_irq, block_writer_select=>block_writer_select,
				tvc_clock=>tvc_clock, tvc_reset=>tvc_reset,
				epp_fb_address=>bw_epp_fb_address, epp_fb_data=>bw_epp_fb_data, epp_start_strobe=>bw_epp_start_strobe, epp_idle=>bw_epp_idle );

	bs_mu : block_set
		port map(mu_buss_address=>mu_buss_address, mu_buss_data=>mu_buss_data, mu_buss_write=>mu_buss_write, mu_buss_strobe=>mu_buss_strobe,
				block_set_irq=>block_set_irq, block_set_select=>block_set_select,
				tvc_clock=>tvc_clock, tvc_reset=>tvc_reset,
				epp_fb_address=>bs_epp_fb_address, epp_num_bytes=>bs_epp_num_bytes, epp_fb_value=>bs_epp_fb_value, epp_start_strobe=>bs_epp_start_strobe, epp_idle=>bs_epp_idle);

	ps_mu : poly_scanline
		port map( mu_buss_address=>mu_buss_address, mu_buss_data=>mu_buss_data, mu_buss_write=>mu_buss_write, mu_buss_strobe=>mu_buss_strobe,		
				poly_scanline_irq=>poly_scanline_irq, poly_scanline_select=>poly_scanline_select, tvc_clock=>tvc_clock, tvc_reset=>tvc_reset,
				epp_fb_address=>ps_epp_fb_address, epp_db_address=>ps_epp_db_address, epp_start_z_depth=>ps_epp_start_z_depth,
				epp_z_increment=>ps_epp_z_increment, epp_scanline_length=>ps_epp_scanline_length, epp_pix_value=>ps_epp_pix_value,
				epp_start_strobe=>ps_epp_start_strobe, epp_idle=>ps_epp_idle);

 
end Behavioral;
